import inspect
import re
from abc import ABC, abstractmethod
from io import StringIO
from typing import Any, Generic, Literal, TypeVar, cast

import joblib
from numpy.typing import ArrayLike, NDArray
from rich.console import Console, Group
from rich.panel import Panel
from rich.tree import Tree
from sklearn.base import BaseEstimator
from sklearn.utils._response import _check_response_method, _get_response_values

from skore._externals._sklearn_compat import is_clusterer
from skore._sklearn.types import PositiveLabel
from skore._utils._measure_time import MeasureTime


class _HelpMixin(ABC):
    """Mixin class providing help for the `help` method and the `__repr__` method."""

    def _get_methods_for_help(self) -> list[tuple[str, Any]]:
        """Get the methods to display in help."""
        methods = inspect.getmembers(self, predicate=inspect.ismethod)
        filtered_methods = []
        for name, method in methods:
            is_private_method = name.startswith("_")
            # we cannot use `isinstance(method, classmethod)` because it is already
            # transformed by the decorator `@classmethod`.
            is_class_method = inspect.ismethod(method) and method.__self__ is type(self)
            is_help_method = name == "help"
            if not (is_private_method or is_class_method or is_help_method):
                filtered_methods.append((name, method))
        return filtered_methods

    def _sort_methods_for_help(
        self, methods: list[tuple[str, Any]]
    ) -> list[tuple[str, Any]]:
        """Sort methods for help display."""
        return sorted(methods)

    def _format_method_name(self, name: str) -> str:
        """Format method name for display."""
        return f"{name}(...)"

    def _get_method_description(self, method: Any) -> str:
        """Get the description for a method."""
        return (
            method.__doc__.split("\n")[0]
            if method.__doc__
            else "No description available"
        )

    def _get_attribute_description(self, name: str) -> str:
        """Get the description of an attribute from its class docstring."""
        if self.__doc__ is None:
            return "No description available"
        # Look for the first sentence on the line below the pattern 'attribute_name : '
        regex_pattern = rf"{name} : .*?\n\s*(.*?)\."
        search_result = re.search(regex_pattern, self.__doc__)
        return search_result.group(1) if search_result else "No description available"

    def _get_help_legend(self) -> str | None:
        """Get the help legend."""
        return None

    @abstractmethod
    def _create_help_tree(self) -> Tree:
        """Create the help tree."""

    @abstractmethod
    def _get_help_panel_title(self) -> str:
        """Get the help panel title."""

    def _create_help_panel(self) -> Panel:
        """Create the help panel."""
        content: Tree | Group
        if self._get_help_legend():
            content = Group(
                self._create_help_tree(),
                f"\n\nLegend:\n{self._get_help_legend()}",
            )
        else:
            content = self._create_help_tree()

        return Panel(
            content,
            title=self._get_help_panel_title(),
            expand=False,
            border_style="orange1",
        )

    def help(self) -> None:
        """Display available methods using rich."""
        from skore import console  # avoid circular import

        console.print(self._create_help_panel())

    def _rich_repr(self, class_name: str) -> str:
        """Return a string representation using rich."""
        string_buffer = StringIO()
        console = Console(file=string_buffer, force_terminal=False)
        console.print(
            Panel(
                "Get guidance using the help() method",
                title=f"[cyan]{class_name}[/cyan]",
                border_style="orange1",
                expand=False,
            )
        )
        return string_buffer.getvalue()


class _BaseReport(_HelpMixin):
    """Base class for all reports."""

    _ACCESSOR_CONFIG: dict[str, dict[str, str]]
    _X_train: ArrayLike | None
    _X_test: ArrayLike | None
    _y_train: ArrayLike | None
    _y_test: ArrayLike | None
    _cache: dict[tuple[Any, ...], Any]
    estimator_: BaseEstimator

    def _get_help_panel_title(self) -> str:
        return ""

    def _get_help_legend(self) -> str:
        return ""

    def _get_attributes_for_help(self) -> list[str]:
        """Get the public attributes to display in help."""
        attributes = []
        xy_attributes = []

        for name in dir(self):
            # Skip private attributes, callables, and accessors
            if (
                name.startswith("_")
                or callable(getattr(self, name))
                or isinstance(getattr(self, name), _BaseAccessor)
            ):
                continue

            # Group X and y attributes separately
            value = getattr(self, name)
            if name.startswith(("X", "y")):
                if value is not None:  # Only include non-None X/y attributes
                    xy_attributes.append(name)
            else:
                attributes.append(name)

        # Sort X/y attributes to keep them grouped
        xy_attributes.sort()
        attributes.sort()

        # Return X/y attributes first, followed by other attributes
        return xy_attributes + attributes

    def _create_help_tree(self) -> Tree:
        """Create a rich Tree with the available tools and accessor methods."""
        tree = Tree(self.__class__.__name__)

        # Add accessor methods first
        for accessor_attr, config in self._ACCESSOR_CONFIG.items():
            accessor = getattr(self, accessor_attr)
            branch = tree.add(f"[bold cyan].{config['name']}[/bold cyan]")

            # Add main accessor methods first
            methods = accessor._get_methods_for_help()
            methods = accessor._sort_methods_for_help(methods)

            # Add methods
            for name, method in methods:
                displayed_name = accessor._format_method_name(name)
                description = accessor._get_method_description(method)
                branch.add(f".{displayed_name} - {description}")

            # Add sub-accessors after main methods
            for sub_attr, sub_obj in inspect.getmembers(accessor):
                if isinstance(sub_obj, _BaseAccessor) and not sub_attr.startswith("_"):
                    sub_branch = branch.add(f"[bold cyan].{sub_attr}[/bold cyan]")

                    # Add sub-accessor methods
                    sub_methods = sub_obj._get_methods_for_help()
                    sub_methods = sub_obj._sort_methods_for_help(sub_methods)

                    for name, method in sub_methods:
                        displayed_name = sub_obj._format_method_name(name)
                        description = sub_obj._get_method_description(method)
                        sub_branch.add(f".{displayed_name.ljust(25)} - {description}")

        # Add base methods
        base_methods = self._get_methods_for_help()
        base_methods = self._sort_methods_for_help(base_methods)

        for name, method in base_methods:
            description = self._get_method_description(method)
            tree.add(f".{name}(...)".ljust(34) + f" - {description}")

        # Add attributes section
        attributes = self._get_attributes_for_help()
        if attributes:
            attr_branch = tree.add("[bold cyan]Attributes[/bold cyan]")
            for attr_name in attributes:
                description = self._get_attribute_description(attr_name)
                attr_branch.add(f".{attr_name.ljust(29)} - {description}")

        return tree


ParentT = TypeVar("ParentT", bound="_BaseReport")


class _BaseAccessor(_HelpMixin, Generic[ParentT]):
    """Base class for all accessors."""

    def __init__(self, parent: ParentT) -> None:
        self._parent = parent

    @abstractmethod
    def _get_help_tree_title(self) -> str:
        """Get the title for the help tree."""
        pass

    def _get_help_panel_title(self) -> str:
        name = self.__class__.__name__.replace("_", "").replace("Accessor", "").lower()
        return f"Available {name} methods"

    def _create_help_tree(self) -> Tree:
        """Create a rich Tree with the available methods."""
        tree = Tree(self._get_help_tree_title())

        methods = self._get_methods_for_help()
        methods = self._sort_methods_for_help(methods)

        for name, method in methods:
            displayed_name = self._format_method_name(name)
            description = self._get_method_description(method)
            tree.add(f".{displayed_name}".ljust(26) + f" - {description}")

        return tree

    def _get_X_y_and_data_source_hash(
        self,
        *,
        data_source: Literal["test", "train", "X_y"],
        X: ArrayLike | None = None,
        y: ArrayLike | None = None,
    ) -> tuple[ArrayLike, ArrayLike | None, int | None]:
        """Get the requested dataset and mention if we should hash before caching.

        Parameters
        ----------
        data_source : {"test", "train", "X_y"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.
            - "X_y" : use the provided `X` and `y` to compute the metric.

        X : array-like of shape (n_samples, n_features) or None, default=None
            The input data.

        y : array-like of shape (n_samples,) or None, default=None
            The target data.

        Returns
        -------
        X : array-like of shape (n_samples, n_features)
            The requested dataset.

        y : array-like of shape (n_samples,)
            The requested dataset.

        data_source_hash : int or None
            The hash of the data source. None when we are able to track the data, and
            thus relying on X_train, y_train, X_test, y_test.
        """
        is_cluster = is_clusterer(self._parent.estimator_)
        if data_source == "test":
            if not (X is None or y is None):
                raise ValueError("X and y must be None when data_source is test.")
            if self._parent._X_test is None or (
                not is_cluster and self._parent._y_test is None
            ):
                missing_data = "X_test" if is_cluster else "X_test and y_test"
                raise ValueError(
                    f"No {data_source} data (i.e. {missing_data}) were provided "
                    f"when creating the report. Please provide the {data_source} "
                    "data either when creating the report or by setting data_source "
                    "to 'X_y' and providing X and y."
                )
            return self._parent._X_test, self._parent._y_test, None
        elif data_source == "train":
            if not (X is None or y is None):
                raise ValueError("X and y must be None when data_source is train.")
            if self._parent._X_train is None or (
                not is_cluster and self._parent._y_train is None
            ):
                missing_data = "X_train" if is_cluster else "X_train and y_train"
                raise ValueError(
                    f"No {data_source} data (i.e. {missing_data}) were provided "
                    f"when creating the report. Please provide the {data_source} "
                    "data either when creating the report or by setting data_source "
                    "to 'X_y' and providing X and y."
                )
            return self._parent._X_train, self._parent._y_train, None
        elif data_source == "X_y":
            if X is None or (not is_cluster and y is None):
                missing_data = "X" if is_cluster else "X and y"
                raise ValueError(
                    f"{missing_data} must be provided when data_source is X_y."
                )
            return X, y, joblib.hash((X, y))
        else:
            raise ValueError(
                f"Invalid data source: {data_source}. Possible values are: "
                "test, train, X_y."
            )


def _get_cached_response_values(
    *,
    cache: dict[tuple[Any, ...], Any],
    estimator_hash: int,
    estimator: BaseEstimator,
    X: ArrayLike | None,
    response_method: str | list[str] | tuple[str, ...],
    pos_label: PositiveLabel | None = None,
    data_source: Literal["test", "train", "X_y"] = "test",
    data_source_hash: int | None = None,
) -> list[tuple[tuple[Any, ...], Any, bool]]:
    """Compute or load from local cache the response values.

    Be aware that the predictions will be loaded from the cache if present, but they
    will not be added to it. The reason is that we want to be able to run this function
    in parallel settings in a thread-safe manner. The update should be done outside of
    this function.

    Parameters
    ----------
    cache : dict
        The cache to use.

    estimator_hash : int
        A hash associated with the estimator such that we can retrieve the data from
        the cache.

    estimator : estimator object
        The estimator.

    X : {array-like, sparse matrix} of shape (n_samples, n_features) or None
        The data.

    response_method : str, list of str or tuple of str
        The response method.

    pos_label : int, float, bool or str, default=None
        The positive label.

    data_source : {"test", "train", "X_y"}, default="test"
        The data source to use.

        - "test" : use the test set provided when creating the report.
        - "train" : use the train set provided when creating the report.
        - "X_y" : use the provided `X` and `y` to compute the metric.

    data_source_hash : int or None
        The hash of the data source when `data_source` is "X_y".

    Returns
    -------
    list of tuples
        A list of tuples, each containing:

        - cache_key : tuple
            The cache key.

        - cache_value : Any
            The cache value. It corresponds to the predictions but also to the predict
            time when it has not been cached yet.

        - is_cached : bool
            Whether the cache value was loaded from the cache.
    """
    prediction_method = _check_response_method(estimator, response_method).__name__

    if data_source == "X_y" and data_source_hash is None:
        # Only trigger hash computation if it was not previously done.
        # If data_source_hash is not None, we internally computed ourself the hash
        # and it is trustful
        data_source_hash = joblib.hash(X)

    if prediction_method not in ("predict_proba", "decision_function"):
        # pos_label is only important in classification and with probabilities
        # and decision functions
        pos_label = None

    cache_key: tuple[Any, ...] = (
        estimator_hash,
        pos_label,
        prediction_method,
        data_source,
        data_source_hash,
    )

    if cache_key in cache:
        cached_predictions = cast(NDArray, cache[cache_key])
        return [(cache_key, cached_predictions, True)]

    with MeasureTime() as predict_time:
        predictions, _ = _get_response_values(
            estimator,
            X=X,
            response_method=prediction_method,
            pos_label=pos_label,
            return_response_method_used=False,
        )

    predict_time_cache_key: tuple[Any, ...] = (
        estimator_hash,
        data_source,
        data_source_hash,
        "predict_time",
    )

    return [
        (cache_key, predictions, False),
        (predict_time_cache_key, predict_time(), False),
    ]


class _BaseMetricsAccessor:
    _score_or_loss_info: dict[str, dict[str, str]] = {
        "fit_time": {"name": "Fit time (s)", "icon": "(↘︎)"},
        "predict_time": {"name": "Predict time (s)", "icon": "(↘︎)"},
        "accuracy": {"name": "Accuracy", "icon": "(↗︎)"},
        "precision": {"name": "Precision", "icon": "(↗︎)"},
        "recall": {"name": "Recall", "icon": "(↗︎)"},
        "brier_score": {"name": "Brier score", "icon": "(↘︎)"},
        "roc_auc": {"name": "ROC AUC", "icon": "(↗︎)"},
        "log_loss": {"name": "Log loss", "icon": "(↘︎)"},
        "r2": {"name": "R²", "icon": "(↗︎)"},
        "rmse": {"name": "RMSE", "icon": "(↘︎)"},
        "custom_metric": {"name": "Custom metric", "icon": ""},
        "report_metrics": {"name": "Report metrics", "icon": ""},
    }

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def _sort_methods_for_help(self, methods: list[tuple]) -> list[tuple]:
        """Override sort method for metrics-specific ordering.

        In short, we display the `summarize` first and then the `custom_metric`.
        """

        def _sort_key(method):
            name = method[0]
            if name == "custom_metric":
                priority = 1
            elif name == "summarize":
                priority = 2
            else:
                priority = 0
            return priority, name

        return sorted(methods, key=_sort_key)

    def _format_method_name(self, name: str) -> str:
        """Override format method for metrics-specific naming."""
        method_name = f"{name}(...)"
        method_name = method_name.ljust(22)
        if name in self._score_or_loss_info and self._score_or_loss_info[name][
            "icon"
        ] in (
            "(↗︎)",
            "(↘︎)",
        ):
            if self._score_or_loss_info[name]["icon"] == "(↗︎)":
                method_name += f"[cyan]{self._score_or_loss_info[name]['icon']}[/cyan]"
                return method_name.ljust(43)
            else:  # (↘︎)
                method_name += (
                    f"[orange1]{self._score_or_loss_info[name]['icon']}[/orange1]"
                )
                return method_name.ljust(49)
        else:
            return method_name.ljust(29)

    def _get_help_panel_title(self) -> str:
        return "[bold cyan]Available metrics methods[/bold cyan]"

    def _get_help_tree_title(self) -> str:
        return "[bold cyan]report.metrics[/bold cyan]"

    def _get_help_legend(self) -> str:
        return (
            "[cyan](↗︎)[/cyan] higher is better [orange1](↘︎)[/orange1] lower is better"
        )



import numbers
from collections import namedtuple
from typing import Any, Literal

import matplotlib.pyplot as plt
import numpy as np
from matplotlib import colormaps
from matplotlib.artist import Artist
from pandas import DataFrame
from sklearn.utils.validation import _num_samples, check_array

from skore._externals._sklearn_compat import _safe_indexing
from skore._sklearn._plot.style import StyleDisplayMixin
from skore._sklearn._plot.utils import (
    HelpDisplayMixin,
    PlotBackendMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)
from skore._sklearn.types import MLTask, ReportType, YPlotData

RangeData = namedtuple("RangeData", ["min", "max"])

MAX_N_LABELS = 6  # 5 + 1 for the perfect model line


class PredictionErrorDisplay(StyleDisplayMixin, HelpDisplayMixin, PlotBackendMixin):
    """Visualization of the prediction error of a regression model.

    This tool can display "residuals vs predicted" or "actual vs predicted"
    using scatter plots to qualitatively assess the behavior of a regressor,
    preferably on held-out data points.

    An instance of this class should be created by
    `EstimatorReport.metrics.prediction_error()`.
    You should not create an instance of this class directly.

    Parameters
    ----------
    prediction_error : DataFrame
        The prediction error data to display. The columns are

        - `estimator_name`
        - `split_index` (may be null)
        - `y_true`
        - `y_pred`
        - `residuals`.

    range_y_true : RangeData
        Global range of the true values.

    range_y_pred : RangeData
        Global range of the predicted values.

    range_residuals : RangeData
        Global range of the residuals.

    data_source : {"train", "test", "X_y"}
        The data source used to display the prediction error.

    ml_task : {"regression", "multioutput-regression"}
        The machine learning task.

    report_type : {"comparison-cross-validation", "comparison-estimator", \
            "cross-validation", "estimator"}
        The type of report.

    Attributes
    ----------
    line_ : matplotlib Artist
        Optimal line representing `y_true == y_pred`. Therefore, it is a
        diagonal line for `kind="predictions"` and a horizontal line for
        `kind="residuals"`.

    errors_lines_ : matplotlib Artist or None
        Residual lines. If `with_errors=False`, then it is set to `None`.

    scatter_ : list of matplotlib Artist
        Scatter data points.

    ax_ : matplotlib Axes
        Axes with the different matplotlib axis.

    figure_ : matplotlib Figure
        Figure containing the scatter and lines.

    Examples
    --------
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.linear_model import Ridge
    >>> from skore import train_test_split
    >>> from skore import EstimatorReport
    >>> X, y = load_diabetes(return_X_y=True)
    >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
    >>> classifier = Ridge()
    >>> report = EstimatorReport(classifier, **split_data)
    >>> display = report.metrics.prediction_error()
    >>> display.plot(kind="actual_vs_predicted")
    """

    _default_data_points_kwargs: dict[str, Any] | None = None
    _default_perfect_model_kwargs: dict[str, Any] | None = None

    def __init__(
        self,
        *,
        prediction_error: DataFrame,
        range_y_true: RangeData,
        range_y_pred: RangeData,
        range_residuals: RangeData,
        data_source: Literal["train", "test", "X_y"],
        ml_task: MLTask,
        report_type: ReportType,
    ) -> None:
        self._prediction_error = prediction_error
        self.range_y_true = range_y_true
        self.range_y_pred = range_y_pred
        self.range_residuals = range_residuals
        self.data_source = data_source
        self.ml_task = ml_task
        self.report_type = report_type

    def _validate_data_points_kwargs(
        self,
        *,
        data_points_kwargs: dict[str, Any] | list[dict[str, Any]] | None,
    ) -> list[dict[str, Any]]:
        """Validate and format the scatter keyword arguments.

        Parameters
        ----------
        data_points_kwargs : dict or list of dict or None
            Keyword arguments for the scatter plot.

        Returns
        -------
        list of dict
            Validated list of keyword arguments for each curve.

        Raises
        ------
        ValueError
            If the format of `data_points_kwargs` is invalid.
        """
        if self.report_type == "estimator":
            allow_single_dict = True
            n_scatter_groups = 1
        elif self.report_type == "cross-validation":
            allow_single_dict = True
            n_scatter_groups = len(self._prediction_error["split_index"].cat.categories)
        elif self.report_type in (
            "comparison-estimator",
            "comparison-cross-validation",
        ):
            # since we compare different estimators, it does not make sense to share
            # a single dictionary for all the estimators.
            allow_single_dict = False
            n_scatter_groups = len(
                self._prediction_error["estimator_name"].cat.categories
            )
        else:
            raise ValueError(
                f"`report_type` should be one of 'estimator', 'cross-validation', "
                "'comparison-cross-validation' or 'comparison-estimator'. "
                f"Got '{self.report_type}' instead."
            )

        if data_points_kwargs is None:
            return [{}] * n_scatter_groups
        elif n_scatter_groups == 1:
            if isinstance(data_points_kwargs, dict):
                return [data_points_kwargs]
            raise ValueError(
                "You intend to plot the prediction error for a single estimator. We "
                "expect `data_points_kwargs` to be a dictionary. Got "
                f"{type(data_points_kwargs)} instead."
            )
        else:
            if not allow_single_dict and isinstance(data_points_kwargs, dict):
                raise ValueError(
                    "You intend to plot multiple curves. We expect "
                    "`data_points_kwargs` to be a list of dictionaries. Got "
                    f"{type(data_points_kwargs)} instead."
                )

            if isinstance(data_points_kwargs, dict):
                return [data_points_kwargs] * n_scatter_groups

            # From this point, we have a list of dictionaries
            if len(data_points_kwargs) != n_scatter_groups:
                raise ValueError(
                    "You intend to plot prediction errors either from multiple "
                    "estimators or from a cross-validated estimator. "
                    "We expect `data_points_kwargs` to be a list of dictionaries "
                    "with the same length as the number of estimators or splits. "
                    f"Got {len(data_points_kwargs)} instead of {n_scatter_groups}."
                )

        return data_points_kwargs

    def _plot_single_estimator(
        self,
        *,
        kind: Literal["actual_vs_predicted", "residual_vs_predicted"],
        estimator_name: str,
        samples_kwargs: list[dict[str, Any]],
    ) -> list[Artist]:
        """Plot the prediction error for a single estimator.

        Parameters
        ----------
        kind : {"actual_vs_predicted", "residual_vs_predicted"}
            The type of plot to draw.

        estimator_name : str
            Name of the estimator.

        samples_kwargs : list of dict
            Keyword arguments for the scatter plot.

        Returns
        -------
        scatter : list of matplotlib Artist
            The scatter plot.
        """
        scatter = []
        data_points_kwargs: dict[str, Any] = {
            "color": "tab:blue",
            "alpha": 0.3,
            "s": 10,
        }

        data_points_kwargs_validated = _validate_style_kwargs(
            data_points_kwargs, samples_kwargs[0]
        )

        if self.data_source in ("train", "test"):
            scatter_label = f"{self.data_source.title()} set"
        else:  # data_source == "X_y"
            scatter_label = "External data set"

        if kind == "actual_vs_predicted":
            scatter.append(
                self.ax_.scatter(
                    self._prediction_error["y_pred"],
                    self._prediction_error["y_true"],
                    label=scatter_label,
                    **data_points_kwargs_validated,
                )
            )
        else:  # kind == "residual_vs_predicted"
            scatter.append(
                self.ax_.scatter(
                    self._prediction_error["y_pred"],
                    self._prediction_error["residuals"],
                    label=scatter_label,
                    **data_points_kwargs_validated,
                )
            )

        # move the perfect model line to the end of the legend
        handles, labels = self.ax_.get_legend_handles_labels()
        handles.append(handles.pop(0))
        labels.append(labels.pop(0))

        self.ax_.legend(handles, labels, loc="lower right")
        self.ax_.set_title(f"Prediction Error for {estimator_name}")

        return scatter

    def _plot_cross_validated_estimator(
        self,
        *,
        kind: Literal["actual_vs_predicted", "residual_vs_predicted"],
        estimator_name: str,
        samples_kwargs: list[dict[str, Any]],
    ) -> list[Artist]:
        """Plot the prediction error for a cross-validated estimator.

        Parameters
        ----------
        kind : {"actual_vs_predicted", "residual_vs_predicted"}
            The type of plot to draw.

        estimator_name : str
            Name of the estimator.

        samples_kwargs : list of dict
            Keyword arguments for the scatter plot.

        Returns
        -------
        scatter : list of matplotlib Artist
            The scatter plot.
        """
        scatter = []
        data_points_kwargs: dict[str, Any] = {"alpha": 0.3, "s": 10}
        n_splits = len(self._prediction_error["split_index"].cat.categories)
        colors_markers = sample_mpl_colormap(
            colormaps.get_cmap("tab10"),
            n_splits if n_splits > 10 else 10,
        )

        for split_idx, prediction_error_split in self._prediction_error.groupby(
            "split_index", observed=True
        ):
            data_points_kwargs_fold = {
                "color": colors_markers[split_idx],
                **data_points_kwargs,
            }

            data_points_kwargs_validated = _validate_style_kwargs(
                data_points_kwargs_fold, samples_kwargs[split_idx]
            )

            label = f"Fold #{split_idx + 1}"

            if kind == "actual_vs_predicted":
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_split["y_pred"],
                        prediction_error_split["y_true"],
                        label=label,
                        **data_points_kwargs_validated,
                    )
                )
            else:  # kind == "residual_vs_predicted"
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_split["y_pred"],
                        prediction_error_split["residuals"],
                        label=label,
                        **data_points_kwargs_validated,
                    )
                )

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        # move the perfect model line to the end of the legend
        handles, labels = self.ax_.get_legend_handles_labels()
        handles.append(handles.pop(0))
        labels.append(labels.pop(0))

        if len(labels) > MAX_N_LABELS or kind == "residual_vs_predicted":
            # too many lines to fit legend in the plot
            self.ax_.legend(
                handles, labels, bbox_to_anchor=(1.02, 1), title=legend_title
            )
        else:
            self.ax_.legend(handles, labels, loc="lower right", title=legend_title)
        self.ax_.set_title(f"Prediction Error for {estimator_name}")

        return scatter

    def _plot_comparison_estimator(
        self,
        *,
        kind: Literal["actual_vs_predicted", "residual_vs_predicted"],
        samples_kwargs: list[dict[str, Any]],
    ) -> list[Artist]:
        """Plot the prediction error of several estimators.

        Parameters
        ----------
        kind : {"actual_vs_predicted", "residual_vs_predicted"}
            The type of plot to draw.

        samples_kwargs : list of dict
            Keyword arguments for the scatter plot.

        Returns
        -------
        scatter : list of matplotlib Artist
            The scatter plot.
        """
        scatter = []
        data_points_kwargs: dict[str, Any] = {"alpha": 0.3, "s": 10}

        estimator_names = self._prediction_error["estimator_name"].cat.categories
        n_estimators = len(estimator_names)
        colors_markers = sample_mpl_colormap(
            colormaps.get_cmap("tab10"),
            n_estimators if n_estimators > 10 else 10,
        )

        for idx, (estimator_name, prediction_error_estimator) in enumerate(
            self._prediction_error.groupby("estimator_name", observed=True)
        ):
            data_points_kwargs_fold = {
                "color": colors_markers[idx],
                **data_points_kwargs,
            }

            data_points_kwargs_validated = _validate_style_kwargs(
                data_points_kwargs_fold, samples_kwargs[idx]
            )

            if kind == "actual_vs_predicted":
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_estimator["y_pred"],
                        prediction_error_estimator["y_true"],
                        label=estimator_name,
                        **data_points_kwargs_validated,
                    )
                )
            else:  # kind == "residual_vs_predicted"
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_estimator["y_pred"],
                        prediction_error_estimator["residuals"],
                        label=estimator_name,
                        **data_points_kwargs_validated,
                    )
                )

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        # move the perfect model line to the end of the legend
        handles, labels = self.ax_.get_legend_handles_labels()
        handles.append(handles.pop(0))
        labels.append(labels.pop(0))

        if len(labels) > MAX_N_LABELS or kind == "residual_vs_predicted":
            # too many lines to fit legend in the plot
            self.ax_.legend(
                handles, labels, bbox_to_anchor=(1.02, 1), title=legend_title
            )
        else:
            self.ax_.legend(handles, labels, loc="lower right", title=legend_title)
        self.ax_.set_title("Prediction Error")

        return scatter

    def _plot_comparison_cross_validation(
        self,
        *,
        kind: Literal["actual_vs_predicted", "residual_vs_predicted"],
        samples_kwargs: list[dict[str, Any]],
    ) -> list[Artist]:
        """Plot the prediction error of several cross-validated estimators.

        Parameters
        ----------
        kind : {"actual_vs_predicted", "residual_vs_predicted"}
            The type of plot to draw.

        samples_kwargs : list of dict
            Keyword arguments for the scatter plot.

        Returns
        -------
        scatter : list of matplotlib Artist
            The scatter plot.
        """
        scatter = []
        data_points_kwargs: dict[str, Any] = {"alpha": 0.3, "s": 10}

        estimator_names = self._prediction_error["estimator_name"].cat.categories
        n_estimators = len(estimator_names)
        colors_markers = sample_mpl_colormap(
            colormaps.get_cmap("tab10"),
            n_estimators if n_estimators > 10 else 10,
        )

        for idx, (estimator_name, prediction_error_estimator) in enumerate(
            self._prediction_error.groupby("estimator_name", observed=True)
        ):
            data_points_kwargs_fold = {
                "color": colors_markers[idx],
                **data_points_kwargs,
            }

            data_points_kwargs_validated = _validate_style_kwargs(
                data_points_kwargs_fold, samples_kwargs[idx]
            )

            if kind == "actual_vs_predicted":
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_estimator["y_pred"],
                        prediction_error_estimator["y_true"],
                        label=estimator_name,
                        **data_points_kwargs_validated,
                    )
                )
            else:  # kind == "residual_vs_predicted"
                scatter.append(
                    self.ax_.scatter(
                        prediction_error_estimator["y_pred"],
                        prediction_error_estimator["residuals"],
                        label=estimator_name,
                        **data_points_kwargs_validated,
                    )
                )

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        # move the perfect model line to the end of the legend
        handles, labels = self.ax_.get_legend_handles_labels()
        handles.append(handles.pop(0))
        labels.append(labels.pop(0))

        if len(labels) > MAX_N_LABELS or kind == "residual_vs_predicted":
            # too many lines to fit legend in the plot
            self.ax_.legend(
                handles, labels, bbox_to_anchor=(1.02, 1), title=legend_title
            )
        else:
            self.ax_.legend(handles, labels, loc="lower right", title=legend_title)
        self.ax_.set_title("Prediction Error")

        return scatter

    @StyleDisplayMixin.style_plot
    def _plot_matplotlib(
        self,
        *,
        estimator_name: str | None = None,
        kind: Literal[
            "actual_vs_predicted", "residual_vs_predicted"
        ] = "residual_vs_predicted",
        data_points_kwargs: dict[str, Any] | list[dict[str, Any]] | None = None,
        perfect_model_kwargs: dict[str, Any] | None = None,
        despine: bool = True,
    ) -> None:
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's ``plot``.

        Parameters
        ----------
        estimator_name : str
            Name of the estimator used to plot the prediction error. If `None`,
            we used the inferred name from the estimator.

        kind : {"actual_vs_predicted", "residual_vs_predicted"}, \
                default="residual_vs_predicted"
            The type of plot to draw:

            - "actual_vs_predicted" draws the observed values (y-axis) vs.
              the predicted values (x-axis).
            - "residual_vs_predicted" draws the residuals, i.e. difference
              between observed and predicted values, (y-axis) vs. the predicted
              values (x-axis).

        data_points_kwargs : dict, default=None
            Dictionary with keywords passed to the `matplotlib.pyplot.scatter`
            call.

        perfect_model_kwargs : dict, default=None
            Dictionary with keyword passed to the `matplotlib.pyplot.plot`
            call to draw the optimal line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_diabetes(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = Ridge()
        >>> report = EstimatorReport(classifier, **split_data)
        >>> display = report.metrics.prediction_error()
        >>> display.plot(kind="actual_vs_predicted")
        """
        expected_kind = ("actual_vs_predicted", "residual_vs_predicted")
        if kind not in expected_kind:
            raise ValueError(
                f"`kind` must be one of {', '.join(expected_kind)}. "
                f"Got {kind!r} instead."
            )
        if kind == "actual_vs_predicted":
            xlabel, ylabel = "Predicted values", "Actual values"
        else:  # kind == "residual_vs_predicted"
            xlabel, ylabel = "Predicted values", "Residuals (actual - predicted)"

        self.figure_, self.ax_ = plt.subplots()

        perfect_model_kwargs_validated = _validate_style_kwargs(
            {
                "color": "black",
                "alpha": 0.7,
                "linestyle": "--",
                "label": "Perfect predictions",
            },
            perfect_model_kwargs or self._default_perfect_model_kwargs or {},
        )

        if kind == "actual_vs_predicted":
            # For actual vs predicted, we want the same range for both axes
            min_value = min(self.range_y_pred.min, self.range_y_true.min)
            max_value = max(self.range_y_pred.max, self.range_y_true.max)
            x_range_perfect_pred = [min_value, max_value]
            y_range_perfect_pred = [min_value, max_value]

            self.line_ = self.ax_.plot(
                x_range_perfect_pred,
                y_range_perfect_pred,
                **perfect_model_kwargs_validated,
            )[0]
            self.ax_.set(
                aspect="equal",
                xlim=x_range_perfect_pred,
                ylim=y_range_perfect_pred,
                xticks=np.linspace(
                    x_range_perfect_pred[0], x_range_perfect_pred[1], num=5
                ),
                yticks=np.linspace(
                    y_range_perfect_pred[0], y_range_perfect_pred[1], num=5
                ),
            )

        else:  # kind == "residual_vs_predicted"
            x_range_perfect_pred = [self.range_y_pred.min, self.range_y_pred.max]
            y_range_perfect_pred = [self.range_residuals.min, self.range_residuals.max]

            self.line_ = self.ax_.plot(
                x_range_perfect_pred, [0, 0], **perfect_model_kwargs_validated
            )[0]
            self.ax_.set(
                xlim=x_range_perfect_pred,
                ylim=y_range_perfect_pred,
                xticks=np.linspace(
                    x_range_perfect_pred[0], x_range_perfect_pred[1], num=5
                ),
                yticks=np.linspace(
                    y_range_perfect_pred[0], y_range_perfect_pred[1], num=5
                ),
            )

        self.ax_.set(xlabel=xlabel, ylabel=ylabel)

        # make the scatter plot afterwards since it should take into account the line
        # for the perfect predictions
        if data_points_kwargs is None:
            data_points_kwargs = self._default_data_points_kwargs
        data_points_kwargs = self._validate_data_points_kwargs(
            data_points_kwargs=data_points_kwargs
        )

        if self.report_type == "estimator":
            self.scatter_ = self._plot_single_estimator(
                kind=kind,
                estimator_name=(
                    self._prediction_error["estimator_name"].cat.categories.item()
                    if estimator_name is None
                    else estimator_name
                ),
                samples_kwargs=data_points_kwargs,
            )
        elif self.report_type == "cross-validation":
            self.scatter_ = self._plot_cross_validated_estimator(
                kind=kind,
                estimator_name=(
                    self._prediction_error["estimator_name"].cat.categories.item()
                    if estimator_name is None
                    else estimator_name
                ),
                samples_kwargs=data_points_kwargs,
            )
        elif self.report_type == "comparison-estimator":
            self.scatter_ = self._plot_comparison_estimator(
                kind=kind,
                samples_kwargs=data_points_kwargs,
            )
        elif self.report_type == "comparison-cross-validation":
            self.scatter_ = self._plot_comparison_cross_validation(
                kind=kind,
                samples_kwargs=data_points_kwargs,
            )
        else:
            raise ValueError(
                f"`report_type` should be one of 'estimator', 'cross-validation', "
                f"or 'comparison-estimator'. Got '{self.report_type}' instead."
            )

        if despine:
            x_range = self.ax_.get_xlim()
            y_range = self.ax_.get_ylim()
            _despine_matplotlib_axis(self.ax_, x_range=x_range, y_range=y_range)

    @classmethod
    def _compute_data_for_display(
        cls,
        y_true: list[YPlotData],
        y_pred: list[YPlotData],
        *,
        report_type: ReportType,
        ml_task: MLTask,
        data_source: Literal["train", "test", "X_y"],
        subsample: float | int | None = 1_000,
        seed: int | None = None,
        **kwargs,
    ) -> "PredictionErrorDisplay":
        """Plot the prediction error given the true and predicted targets.

        Parameters
        ----------
        y_true : list of array-like of shape (n_samples,)
            True target values.

        y_pred : list of array-like of shape (n_samples,)
            Predicted target values.

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        estimators : list of estimator instances
            The estimators from which `y_pred` is obtained.

        ml_task : {"regression", "multioutput-regression"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}
            The data source used to compute the prediction error curve.

        subsample : float, int or None, default=1_000
            Sampling the samples to be shown on the scatter plot. If `float`,
            it should be between 0 and 1 and represents the proportion of the
            original dataset. If `int`, it represents the number of samples
            display on the scatter plot. If `None`, no subsampling will be
            applied. by default, 1000 samples or less will be displayed.

        seed : int, default=None
            The seed used to initialize the random number generator used for the
            subsampling.

        **kwargs : dict
            Additional keyword arguments that are ignored for compatibility with
            other metrics displays. Here, `estimators` is ignored.

        Returns
        -------
        display : PredictionErrorDisplay
        """
        rng = np.random.default_rng(seed)
        if isinstance(subsample, numbers.Integral):
            if subsample <= 0:
                raise ValueError(
                    f"When an integer, subsample={subsample} should be positive."
                )
        elif isinstance(subsample, numbers.Real) and (subsample <= 0 or subsample >= 1):
            raise ValueError(
                f"When a floating-point, subsample={subsample} should be in the "
                "(0, 1) range."
            )

        if ml_task != "regression":  # pragma: no cover
            raise ValueError(
                "The machine learning task must be 'regression'. "
                f"Got {ml_task} instead."
            )

        prediction_error_records = []
        y_true_min, y_true_max = np.inf, -np.inf
        y_pred_min, y_pred_max = np.inf, -np.inf
        residuals_min, residuals_max = np.inf, -np.inf

        for y_true_i, y_pred_i in zip(y_true, y_pred, strict=False):
            n_samples = _num_samples(y_true_i.y)
            if subsample is None:
                subsample_ = n_samples
            elif isinstance(subsample, numbers.Integral):
                subsample_ = subsample
            else:  # subsample is a float
                subsample_ = int(n_samples * subsample)

            # normalize subsample based on the number of splits
            subsample_ = int(subsample_ / len(y_true))
            if subsample_ < n_samples:
                indices = rng.choice(np.arange(n_samples), size=subsample_)
                y_true_sample = check_array(
                    _safe_indexing(y_true_i.y, indices, axis=0), ensure_2d=False
                )
                y_pred_sample = check_array(
                    _safe_indexing(y_pred_i.y, indices, axis=0), ensure_2d=False
                )
                residuals_sample = y_true_sample - y_pred_sample

                for y_true_sample_i, y_pred_sample_i, residuals_sample_i in zip(
                    y_true_sample, y_pred_sample, residuals_sample, strict=False
                ):
                    prediction_error_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "y_true": y_true_sample_i,
                            "y_pred": y_pred_sample_i,
                            "residuals": residuals_sample_i,
                        }
                    )
            else:
                y_true_sample = y_true_i.y
                y_pred_sample = y_pred_i.y
                residuals_sample = y_true_i.y - y_pred_i.y

                for y_true_sample_i, y_pred_sample_i, residuals_sample_i in zip(
                    y_true_sample, y_pred_sample, residuals_sample, strict=False
                ):
                    prediction_error_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "y_true": y_true_sample_i,
                            "y_pred": y_pred_sample_i,
                            "residuals": residuals_sample_i,
                        }
                    )

            y_true_min = min(y_true_min, np.min(y_true_sample))
            y_true_max = max(y_true_max, np.max(y_true_sample))
            y_pred_min = min(y_pred_min, np.min(y_pred_sample))
            y_pred_max = max(y_pred_max, np.max(y_pred_sample))
            residuals_min = min(residuals_min, np.min(residuals_sample))
            residuals_max = max(residuals_max, np.max(residuals_sample))

        range_y_true = RangeData(min=y_true_min, max=y_true_max)
        range_y_pred = RangeData(min=y_pred_min, max=y_pred_max)
        range_residuals = RangeData(min=residuals_min, max=residuals_max)

        return cls(
            prediction_error=DataFrame.from_records(prediction_error_records).astype(
                {"estimator_name": "category", "split_index": "category"}
            ),
            range_y_true=range_y_true,
            range_y_pred=range_y_pred,
            range_residuals=range_residuals,
            data_source=data_source,
            ml_task=ml_task,
            report_type=report_type,
        )

    def frame(self) -> DataFrame:
        """Get the data used to create the prediction error plot.

        Returns
        -------
        DataFrame
            A DataFrame containing the prediction error data with columns depending on
            the report type:

            - `estimator_name`: Name of the estimator (when comparing estimators)
            - `split_index`: Cross-validation fold ID (when doing cross-validation)
            - `y_true`: True target values
            - `y_pred`: Predicted target values
            - `residuals`: Difference between true and predicted values
              `(y_true - y_pred)`


        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split, EstimatorReport
        >>> X, y = load_diabetes(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> reg = Ridge()
        >>> report = EstimatorReport(reg, **split_data)
        >>> display = report.metrics.prediction_error()
        >>> df = display.frame()
        """
        statistical_columns = ["y_true", "y_pred", "residuals"]
        if self.report_type == "estimator":
            columns = statistical_columns
        elif self.report_type == "cross-validation":
            columns = ["split_index"] + statistical_columns
        elif self.report_type == "comparison-estimator":
            columns = ["estimator_name"] + statistical_columns
        else:  # self.report_type == "comparison-cross-validation"
            columns = ["estimator_name", "split_index"] + statistical_columns

        return self._prediction_error[columns]



"""Global configuration state and functions for management."""

from __future__ import annotations

import threading
import time
from collections.abc import Generator
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from skore._sklearn.types import PlotBackend

_global_config: dict[str, Any] = {
    "show_progress": True,
    "plot_backend": "matplotlib",
}
_threadlocal = threading.local()


def _get_threadlocal_config() -> dict[str, Any]:
    """Get a threadlocal **mutable** configuration.

    If the configuration does not exist, copy the default global configuration.
    """
    if not hasattr(_threadlocal, "global_config"):
        _threadlocal.global_config = _global_config.copy()
    return _threadlocal.global_config


def get_config() -> dict[str, Any]:
    """Retrieve current values for configuration set by :func:`set_config`.

    Returns
    -------
    config : dict
        Keys are parameter names that can be passed to :func:`set_config`.

    See Also
    --------
    config_context : Context manager for skore configuration.
    set_config : Set skore configuration.

    Examples
    --------
    >>> import skore
    >>> config = skore.get_config()
    >>> config.keys()
    dict_keys([...])
    """
    # Return a copy of the threadlocal configuration so that users will
    # not be able to modify the configuration with the returned dict.
    return _get_threadlocal_config().copy()


def set_config(
    *,
    show_progress: bool | None = None,
    plot_backend: PlotBackend | None = None,
) -> None:
    """Set skore configuration.

    Setting the configuration affects global settings meaning that it will be used
    by all skore functions and classes, even in the processes and threads spawned by
    skore.

    Parameters
    ----------
    show_progress : bool, default=None
        If True, show progress bars. Otherwise, do not show them.

    plot_backend : {"matplotlib", "plotly"}, default=None
        The plotting backend to be used.

        - `"matplotlib"`: Use Matplotlib for plotting
        - `"plotly"`: Use Plotly for plotting
        - `None`: Plotting backend is unchanged

    See Also
    --------
    config_context : Context manager for skore configuration.
    get_config : Retrieve current values of the configuration.

    Examples
    --------
    >>> # xdoctest: +SKIP
    >>> from skore import set_config
    >>> set_config(show_progress=False, plot_backend="plotly")
    """
    local_config = _get_threadlocal_config()

    if show_progress is not None:
        local_config["show_progress"] = show_progress
    if plot_backend is not None:
        local_config["plot_backend"] = plot_backend


@contextmanager
def config_context(
    *,
    show_progress: bool | None = None,
    plot_backend: PlotBackend | None = None,
) -> Generator[None, None, None]:
    """Context manager for skore configuration.

    Setting the configuration affects global settings meaning that it will be used
    by all skore functions and classes, even in the processes and threads spawned by
    skore.

    Parameters
    ----------
    show_progress : bool, default=None
        If True, show progress bars. Otherwise, do not show them.

    plot_backend : {"matplotlib", "plotly"}, default=None
        The plotting backend to be used.

        - `"matplotlib"`: Use Matplotlib for plotting
        - `"plotly"`: Use Plotly for plotting
        - `None`: Plotting backend is unchanged

    Yields
    ------
    None.

    See Also
    --------
    set_config : Set skore configuration.
    get_config : Retrieve current values of the configuration.

    Notes
    -----
    All settings, not just those presently modified, will be returned to
    their previous values when the context manager is exited.

    Examples
    --------
    >>> import skore
    >>> from skore import train_test_split
    >>> from sklearn.datasets import make_classification
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import CrossValidationReport
    >>> with skore.config_context(show_progress=False, plot_backend="matplotlib"):
    ...     X, y = make_classification(random_state=42)
    ...     estimator = LogisticRegression()
    ...     report = CrossValidationReport(estimator, X=X, y=y, splitter=2)
    """
    old_config = get_config()
    set_config(
        show_progress=show_progress,
        plot_backend=plot_backend,
    )

    try:
        yield
    finally:
        set_config(**old_config)


def _set_show_progress_for_testing(show_progress: bool, sleep_duration: float) -> bool:
    """Set the value of show_progress for testing purposes after some waiting.

    This function should exist in a Python module rather than in tests, otherwise
    joblib will not be able to pickle it.
    """
    with config_context(show_progress=show_progress):
        time.sleep(sleep_duration)
        return get_config()["show_progress"]



"""Pandas-like accessors.

This code is copied from:
https://github.com/pandas-dev/pandas/blob/main/pandas/core/accessor.py

It is used to register accessors for the skore classes.
"""

from typing import final


class DirNamesMixin:
    _accessors: set[str] = set()
    _hidden_attrs: frozenset[str] = frozenset()

    @final
    def _dir_deletions(self) -> set[str]:
        return self._accessors | self._hidden_attrs

    def _dir_additions(self) -> set[str]:
        return {accessor for accessor in self._accessors if hasattr(self, accessor)}

    def __dir__(self) -> list[str]:
        rv = set(super().__dir__())
        rv = (rv - self._dir_deletions()) | self._dir_additions()
        return sorted(rv)


class Accessor:
    def __init__(self, name: str, accessor) -> None:
        self._name = name
        self._accessor = accessor

    def __get__(self, obj, cls):
        if obj is None:
            # we're accessing the attribute of the class, i.e., Dataset.geo
            return self._accessor
        return self._accessor(obj)


def _register_accessor(name, cls):
    def decorator(accessor):
        if hasattr(cls, name):
            raise ValueError(
                f"registration of accessor {accessor!r} under name "
                f"{name!r} for type {cls!r} is overriding a preexisting "
                f"attribute with the same name."
            )
        setattr(cls, name, Accessor(name, accessor))
        cls._accessors.add(name)
        return accessor

    return decorator



"""Class definition of the ``skore`` project."""

from __future__ import annotations

import re
from importlib.metadata import entry_points
from typing import Any

from skore._sklearn._estimator.report import EstimatorReport
from skore.project.summary import Summary


class Project:
    r"""
    API to manage a collection of key-report pairs.

    Its constructor initializes a project by creating a new project or by loading an
    existing one.

    The class main methods are :func:`~skore.Project.put`,
    :func:`~skore.Project.summarize` and :func:`~skore.Project.get`, respectively to
    insert a key-report pair into the project, to obtain the metadata/metrics of the
    inserted reports and to get a specific report by its id.

    Two mutually exclusive modes are available and can be configured using the ``name``
    parameter of the constructor:

    .. rubric:: Hub mode

    If the ``name`` takes the form of the URI ``hub://<tenant>/<name>``, the project
    is configured to the ``hub`` mode to communicate with the ``skore hub``.

    A tenant is a ``skore hub`` concept that must be configured on the ``skore hub``
    interface. It represents an isolated entity managing users, projects, and
    resources. It can be a company, organization, or team that operates
    independently within the system.

    In this mode, you must have an account to the ``skore hub`` and must be
    authorized to the specified tenant. You must also be authenticated beforehand,
    using the ``skore-hub-login`` CLI.

    .. rubric:: Local mode

    Otherwise, the project is configured to the ``local`` mode to be persisted on
    the user machine in a directory called ``workspace``.

    | The workspace can be shared between all the projects.
    | The workspace can be set using kwargs or the environment variable
      ``SKORE_WORKSPACE``.
    | If not, it will be by default set to a ``skore/`` directory in the USER
      cache directory:

    - on Windows, usually ``C:\Users\%USER%\AppData\Local\skore``,
    - on Linux, usually ``${HOME}/.cache/skore``,
    - on macOS, usually ``${HOME}/Library/Caches/skore``.

    Refer to the :ref:`project` section of the user guide for more details.

    Parameters
    ----------
    name : str
        The name of the project:

        - if the ``name`` takes the form of the URI ``hub://<tenant>/<name>``, the
          project is configured to communicate with the ``skore hub``,
        - otherwise, the project is configured to communicate with a local storage, on
          the user machine.
    **kwargs : dict
        Extra keyword arguments passed to the project, depending on its mode.

        workspace : Path, mode:local only.
            The directory where the local project is persisted.

            | The workspace can be shared between all the projects.
            | The workspace can be set using kwargs or the environment variable
              ``SKORE_WORKSPACE``.
            | If not, it will be by default set to a ``skore/`` directory in the USER
              cache directory:

            - on Windows, usually ``C:\Users\%USER%\AppData\Local\skore``,
            - on Linux, usually ``${HOME}/.cache/skore``,
            - on macOS, usually ``${HOME}/Library/Caches/skore``.

    Attributes
    ----------
    name : str
        The name of the project, extrapolated from the ``name`` parameter.
    mode : str
        The mode of the project, extrapolated from the ``name`` parameter.

    Examples
    --------
    Construct reports.

    >>> from sklearn.datasets import make_classification, make_regression
    >>> from sklearn.linear_model import LinearRegression, LogisticRegression
    >>> from sklearn.model_selection import train_test_split
    >>> from skore._sklearn import EstimatorReport
    >>>
    >>> X, y = make_classification(random_state=42)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
    >>> classifier = LogisticRegression(max_iter=10)
    >>> classifier_report = EstimatorReport(
    >>>     classifier,
    >>>     X_train=X_train,
    >>>     y_train=y_train,
    >>>     X_test=X_test,
    >>>     y_test=y_test,
    >>> )
    >>>
    >>> X, y = make_regression(random_state=42)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y, random_state=42)
    >>> regressor = LinearRegression()
    >>> regressor_report = EstimatorReport(
    >>>     regressor,
    >>>     X_train=X_train,
    >>>     y_train=y_train,
    >>>     X_test=X_test,
    >>>     y_test=y_test,
    >>> )

    Construct the project in local mode, persisted in a temporary directory.

    >>> from pathlib import Path
    >>> from tempfile import TemporaryDirectory
    >>> from skore import Project
    >>>
    >>> tmpdir = TemporaryDirectory().name
    >>> local_project = Project("my-xp", workspace=Path(tmpdir))

    Put reports in the project.

    >>> local_project.put("my-simple-classification", classifier_report)
    >>> local_project.put("my-simple-regression", regressor_report)

    Investigate metadata/metrics to filter the best reports.

    >>> summary = local_project.summarize()
    >>> summary = summary.query("ml_task.str.contains('regression') and (rmse < 67)")
    >>> reports = summary.reports()

    See Also
    --------
    :class:`~skore.project.summary.Summary` :
        DataFrame designed to investigate persisted reports' metadata/metrics.
    """

    __HUB_NAME_PATTERN = re.compile(r"hub://(?P<tenant>[^/]+)/(?P<name>.+)")

    @staticmethod
    def __setup_plugin(name: str) -> tuple[str, str, Any, dict]:
        if not (PLUGINS := entry_points(group="skore.plugins.project")):
            raise SystemError("No project plugin found, please install at least one.")

        if match := re.match(Project.__HUB_NAME_PATTERN, name):
            mode = "hub"
            name = match["name"]
            parameters = {"tenant": match["tenant"], "name": name}
        else:
            mode = "local"
            parameters = {"name": name}

        if mode not in PLUGINS.names:
            raise ValueError(
                f"Unknown mode `{mode}`. "
                f"Please install the `skore-{mode}-project` python package."
            )

        return mode, name, PLUGINS[mode].load(), parameters

    def __init__(self, name: str, **kwargs):
        r"""
        Initialize a project.

        Parameters
        ----------
        name : str
            The name of the project:

            - if the ``name`` takes the form of the URI ``hub://<tenant>/<name>``, the
              project is configured to communicate with the ``skore hub``,
            - otherwise, the project is configured to communicate with a local storage,
              on the user machine.
        **kwargs : dict
            Extra keyword arguments passed to the project, depending on its mode.

            workspace : Path, mode:local only.
                The directory where the local project is persisted.

                | The workspace can be shared between all the projects.
                | The workspace can be set using kwargs or the environment variable
                  ``SKORE_WORKSPACE``.
                | If not, it will be by default set to a ``skore/`` directory in the
                  USER cache directory:

                - on Windows, usually ``C:\Users\%USER%\AppData\Local\skore``,
                - on Linux, usually ``${HOME}/.cache/skore``,
                - on macOS, usually ``${HOME}/Library/Caches/skore``.
        """
        mode, name, plugin, parameters = Project.__setup_plugin(name)

        self.__mode = mode
        self.__name = name
        self.__project = plugin(**(kwargs | parameters))

    @property
    def mode(self):
        """The mode of the project."""
        return self.__mode

    @property
    def name(self):
        """The name of the project."""
        return self.__name

    def put(self, key: str, report: EstimatorReport):
        """
        Put a key-report pair to the project.

        If the key already exists, its last report is modified to point to this new
        report, while keeping track of the report history.

        Parameters
        ----------
        key : str
            The key to associate with ``report`` in the project.
        report : skore.EstimatorReport
            The report to associate with ``key`` in the project.

        Raises
        ------
        TypeError
            If the combination of parameters are not valid.
        """
        if not isinstance(key, str):
            raise TypeError(f"Key must be a string (found '{type(key)}')")

        if not isinstance(report, EstimatorReport):
            raise TypeError(
                f"Report must be a `skore.EstimatorReport` (found '{type(report)}')"
            )

        return self.__project.put(key=key, report=report)

    def get(self, id: str) -> EstimatorReport:
        """
        Get a persisted report by its id.

        Report IDs can be found via :meth:`skore.Project.summarize`, which is
        also the preferred method of interacting with a ``skore.Project``.

        Parameters
        ----------
        id : str
            The id of a report already put in the ``project``.

        Raises
        ------
        KeyError
            If a non-existent ID is passed.
        """
        return self.__project.reports.get(id)

    def summarize(self) -> Summary:
        """Obtain metadata/metrics for all persisted reports."""
        return Summary.factory(self.__project)

    def __repr__(self) -> str:  # noqa: D105
        return self.__project.__repr__()

    @staticmethod
    def delete(name: str, **kwargs):
        r"""
        Delete a project.

        Parameters
        ----------
        name : str
            The name of the project:

            - if the ``name`` takes the form of the URI ``hub://<tenant>/<name>``, the
              project is configured to communicate with the ``skore hub``,
            - otherwise, the project is configured to communicate with a local storage,
              on the user machine.
        **kwargs : dict
            Extra keyword arguments passed to the project, depending on its mode.

            workspace : Path, mode:local only.
                The directory where the local project is persisted.

                | The workspace can be shared between all the projects.
                | The workspace can be set using kwargs or the environment variable
                  ``SKORE_WORKSPACE``.
                | If not, it will be by default set to a ``skore/`` directory in the
                  USER cache directory:

                - on Windows, usually ``C:\Users\%USER%\AppData\Local\skore``,
                - on Linux, usually ``${HOME}/.cache/skore``,
                - on macOS, usually ``${HOME}/Library/Caches/skore``.
        """
        _, _, plugin, parameters = Project.__setup_plugin(name)

        return plugin.delete(**(kwargs | parameters))



from collections.abc import Sequence
from typing import Any, Literal, cast

import matplotlib.pyplot as plt
from matplotlib import colormaps
from matplotlib.axes import Axes
from matplotlib.lines import Line2D
from numpy.typing import NDArray
from pandas import DataFrame
from sklearn.base import BaseEstimator
from sklearn.metrics import auc, roc_curve
from sklearn.preprocessing import LabelBinarizer

from skore._sklearn._plot.style import StyleDisplayMixin
from skore._sklearn._plot.utils import (
    LINESTYLE,
    HelpDisplayMixin,
    PlotBackendMixin,
    _ClassifierCurveDisplayMixin,
    _despine_matplotlib_axis,
    _validate_style_kwargs,
    sample_mpl_colormap,
)
from skore._sklearn.types import MLTask, PositiveLabel, ReportType, YPlotData

MAX_N_LABELS = 6  # 5 + 1 for the chance level line


def _set_axis_labels(ax: Axes, info_pos_label: str | None) -> None:
    """Add axis labels."""
    xlabel = "False Positive Rate"
    ylabel = "True Positive Rate"
    if info_pos_label:
        xlabel += info_pos_label
        ylabel += info_pos_label

    ax.set(
        xlabel=xlabel,
        xlim=(-0.01, 1.01),
        ylabel=ylabel,
        ylim=(-0.01, 1.01),
        aspect="equal",
    )


def _add_chance_level(
    ax: Axes,
    chance_level_kwargs: dict | None,
    default_chance_level_kwargs: dict | None,
) -> Line2D:
    """Add the chance-level line."""
    chance_level_kwargs = _validate_style_kwargs(
        {
            "label": "Chance level (AUC = 0.5)",
            "color": "k",
            "linestyle": "--",
        },
        chance_level_kwargs or default_chance_level_kwargs or {},
    )

    (chance_level,) = ax.plot((0, 1), (0, 1), **chance_level_kwargs)

    return cast(Line2D, chance_level)


class RocCurveDisplay(
    StyleDisplayMixin, HelpDisplayMixin, _ClassifierCurveDisplayMixin, PlotBackendMixin
):
    """ROC Curve visualization.

    An instance of this class should be created by `EstimatorReport.metrics.roc()`.
    You should not create an instance of this class directly.

    Parameters
    ----------
    roc_curve : DataFrame
        The ROC curve data to display. The columns are

        - `estimator_name`
        - `split_index` (may be null)
        - `label`
        - `threshold`
        - `fpr`
        - `tpr`.

    roc_auc : DataFrame
        The ROC AUC data to display. The columns are

        - `estimator_name`
        - `split_index` (may be null)
        - `label`
        - `roc_auc`.

    pos_label : int, float, bool, str or None
        The class considered as positive. Only meaningful for binary classification.

    data_source : {"train", "test", "X_y"}
        The data source used to compute the ROC curve.

    ml_task : {"binary-classification", "multiclass-classification"}
        The machine learning task.

    report_type : {"comparison-cross-validation", "comparison-estimator", \
            "cross-validation", "estimator"}
        The type of report.

    Attributes
    ----------
    ax_ : matplotlib axes or array of axes
        The axes on which the ROC curve is plotted.

    figure_ : matplotlib figure
        The figure on which the ROC curve is plotted.

    lines_ : list of matplotlib lines
        The lines of the ROC curve.

    chance_level_ : matplotlib line or list of lines or None
        The chance level line.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import train_test_split
    >>> from skore import EstimatorReport
    >>> X, y = load_breast_cancer(return_X_y=True)
    >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
    >>> classifier = LogisticRegression(max_iter=10_000)
    >>> report = EstimatorReport(classifier, **split_data)
    >>> display = report.metrics.roc()
    >>> display.plot(roc_curve_kwargs={"color": "tab:red"})
    """

    _default_roc_curve_kwargs: dict[str, Any] | None = None
    _default_chance_level_kwargs: dict[str, Any] | None = None

    def __init__(
        self,
        *,
        roc_curve: DataFrame,
        roc_auc: DataFrame,
        pos_label: PositiveLabel | None,
        data_source: Literal["train", "test", "X_y"],
        ml_task: MLTask,
        report_type: ReportType,
    ) -> None:
        self.roc_curve = roc_curve
        self.roc_auc = roc_auc
        self.pos_label = pos_label
        self.data_source = data_source
        self.ml_task = ml_task
        self.report_type = report_type

    def _plot_single_estimator(
        self,
        *,
        estimator_name: str,
        roc_curve_kwargs: list[dict[str, Any]],
        plot_chance_level: bool = True,
        chance_level_kwargs: dict[str, Any] | None,
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot ROC curve for a single estimator.

        Parameters
        ----------
        estimator_name : str
            The name of the estimator.

        roc_curve_kwargs : list of dict
            Additional keyword arguments to pass to matplotlib's plot function. In
            binary case, we should have a single dict. In multiclass case, we should
            have a list of dicts, one per class.

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes containing the plot.

        lines : list of matplotlib.lines.Line2D
            The plotted lines.

        info_pos_label : str or None
            String containing the positive label info for binary classification.
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {}

        if self.ml_task == "binary-classification":
            if self.data_source in ("train", "test"):
                line_kwargs["label"] = (
                    f"{self.data_source.title()} set "
                    f"(AUC = {self.roc_auc['roc_auc'].item():0.2f})"
                )
            else:  # data_source in (None, "X_y")
                line_kwargs["label"] = f"AUC = {self.roc_auc['roc_auc'].item():0.2f}"

            line_kwargs_validated = _validate_style_kwargs(
                line_kwargs, roc_curve_kwargs[0]
            )

            (line,) = self.ax_.plot(
                self.roc_curve["fpr"],
                self.roc_curve["tpr"],
                **line_kwargs_validated,
            )
            lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
            legend_title = None

        else:  # multiclass-classification
            labels = self.roc_curve["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"), 10 if len(labels) < 10 else len(labels)
            )

            for class_idx, class_label in enumerate(labels):
                query = f"label == {class_label}"
                roc_curve = self.roc_curve.query(query)
                roc_auc = self.roc_auc.query(query)["roc_auc"].item()

                roc_curve_kwargs_class = roc_curve_kwargs[class_idx]

                default_line_kwargs: dict[str, Any] = {"color": class_colors[class_idx]}
                default_line_kwargs["label"] = (
                    f"{str(class_label).title()} (AUC = {roc_auc:0.2f})"
                )

                line_kwargs = _validate_style_kwargs(
                    default_line_kwargs, roc_curve_kwargs_class
                )

                (line,) = self.ax_.plot(
                    roc_curve["fpr"],
                    roc_curve["tpr"],
                    **line_kwargs,
                )
                lines.append(line)

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = None
            info_pos_label = None  # irrelevant for multiclass

        if plot_chance_level:
            self.chance_level_ = _add_chance_level(
                self.ax_,
                chance_level_kwargs,
                self._default_chance_level_kwargs,
            )
        else:
            self.chance_level_ = None

        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower right", title=legend_title)
        self.ax_.set_title(f"ROC Curve for {estimator_name}")

        return self.ax_, lines, info_pos_label

    def _plot_cross_validated_estimator(
        self,
        *,
        estimator_name: str,
        roc_curve_kwargs: list[dict[str, Any]],
        plot_chance_level: bool = True,
        chance_level_kwargs: dict[str, Any] | None,
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot ROC curve for a cross-validated estimator.

        Parameters
        ----------
        estimator_name : str
            The name of the estimator.

        roc_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the ROC
            curves. The length of the list should match the number of curves to plot.

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the ROC curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted ROC curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {}

        if self.ml_task == "binary-classification":
            for split_idx in self.roc_curve["split_index"].cat.categories:
                query = f"label == {self.pos_label!r} & split_index == {split_idx}"
                roc_curve = self.roc_curve.query(query)
                roc_auc = self.roc_auc.query(query)["roc_auc"].item()

                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, roc_curve_kwargs[split_idx]
                )
                line_kwargs_validated["label"] = (
                    f"Fold #{split_idx + 1} (AUC = {roc_auc:0.2f})"
                )

                (line,) = self.ax_.plot(
                    roc_curve["fpr"],
                    roc_curve["tpr"],
                    **line_kwargs_validated,
                )
                lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.roc_curve["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"), 10 if len(labels) < 10 else len(labels)
            )

            for class_idx, class_label in enumerate(labels):
                roc_auc = self.roc_auc.query(f"label == {class_label}")["roc_auc"]
                roc_curve_kwargs_class = roc_curve_kwargs[class_idx]

                for split_idx in self.roc_curve["split_index"].cat.categories:
                    roc_curve_label = self.roc_curve.query(
                        f"label == {class_label} & split_index == {split_idx}"
                    )

                    line_kwargs_validated = _validate_style_kwargs(
                        {
                            "color": class_colors[class_idx],
                            "alpha": 0.3,
                        },
                        roc_curve_kwargs_class,
                    )
                    if split_idx == 0:
                        line_kwargs_validated["label"] = (
                            f"{str(class_label).title()} "
                            f"(AUC = {roc_auc.mean():0.2f} +/- "
                            f"{roc_auc.std():0.2f})"
                        )
                    else:
                        line_kwargs_validated["label"] = None

                    (line,) = self.ax_.plot(
                        roc_curve_label["fpr"],
                        roc_curve_label["tpr"],
                        **line_kwargs_validated,
                    )
                    lines.append(line)

        if plot_chance_level:
            self.chance_level_ = _add_chance_level(
                self.ax_,
                chance_level_kwargs,
                self._default_chance_level_kwargs,
            )
        else:
            self.chance_level_ = None

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower right", title=legend_title)
        self.ax_.set_title(f"ROC Curve for {estimator_name}")

        return self.ax_, lines, info_pos_label

    def _plot_comparison_estimator(
        self,
        *,
        estimator_names: list[str],
        roc_curve_kwargs: list[dict[str, Any]],
        plot_chance_level: bool = True,
        chance_level_kwargs: dict[str, Any] | None,
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot ROC curve of several estimators.

        Parameters
        ----------
        estimator_names : list of str
            The names of the estimators.

        roc_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the ROC
            curves. The length of the list should match the number of curves to plot.

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the ROC curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted ROC curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {}

        if self.ml_task == "binary-classification":
            for est_idx, est_name in enumerate(estimator_names):
                query = f"label == {self.pos_label!r} & estimator_name == '{est_name}'"

                roc_curve = self.roc_curve.query(query)

                roc_auc = self.roc_auc.query(query)["roc_auc"].item()

                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, roc_curve_kwargs[est_idx]
                )
                line_kwargs_validated["label"] = f"{est_name} (AUC = {roc_auc:0.2f})"
                (line,) = self.ax_.plot(
                    roc_curve["fpr"],
                    roc_curve["tpr"],
                    **line_kwargs_validated,
                )
                lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )
        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.roc_curve["label"].cat.categories
            class_colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"), 10 if len(labels) < 10 else len(labels)
            )

            for est_idx, est_name in enumerate(estimator_names):
                est_color = class_colors[est_idx]

                for class_idx, class_label in enumerate(labels):
                    query = f"label == {class_label} & estimator_name == '{est_name}'"
                    roc_curve = self.roc_curve.query(query)

                    roc_auc = self.roc_auc.query(query)["roc_auc"].item()

                    class_linestyle = LINESTYLE[(class_idx % len(LINESTYLE))][1]

                    line_kwargs["color"] = est_color
                    line_kwargs["alpha"] = 0.6
                    line_kwargs["linestyle"] = class_linestyle

                    line_kwargs_validated = _validate_style_kwargs(
                        line_kwargs, roc_curve_kwargs[est_idx]
                    )
                    line_kwargs_validated["label"] = (
                        f"{est_name} - {str(class_label).title()} "
                        f"(AUC = {roc_auc:0.2f})"
                    )

                    (line,) = self.ax_.plot(
                        roc_curve["fpr"], roc_curve["tpr"], **line_kwargs_validated
                    )
                    lines.append(line)

        if plot_chance_level:
            self.chance_level_ = _add_chance_level(
                self.ax_,
                chance_level_kwargs,
                self._default_chance_level_kwargs,
            )
        else:
            self.chance_level_ = None

        if self.data_source in ("train", "test"):
            legend_title = f"{self.data_source.capitalize()} set"
        else:
            legend_title = "External set"

        _, labels = self.ax_.get_legend_handles_labels()
        if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
            self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
        else:
            self.ax_.legend(loc="lower right", title=legend_title)
        self.ax_.set_title("ROC Curve")

        return self.ax_, lines, info_pos_label

    def _plot_comparison_cross_validation(
        self,
        *,
        estimator_names: list[str],
        roc_curve_kwargs: list[dict[str, Any]],
        plot_chance_level: bool = True,
        chance_level_kwargs: dict[str, Any] | None,
    ) -> tuple[Axes, list[Line2D], str | None]:
        """Plot ROC curve of several cross-validations.

        Parameters
        ----------
        estimator_names : list of str
            The names of the estimators.

        roc_curve_kwargs : list of dict
            List of dictionaries containing keyword arguments to customize the ROC
            curves. The length of the list should match the number of curves to plot.

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        Returns
        -------
        ax : matplotlib.axes.Axes
            The axes with the ROC curves plotted.

        lines : list of matplotlib.lines.Line2D
            The plotted ROC curve lines.

        info_pos_label : str or None
            String containing positive label information for binary classification,
            None for multiclass.
        """
        lines: list[Line2D] = []
        line_kwargs: dict[str, Any] = {}

        if self.ml_task == "binary-classification":
            labels = self.roc_curve["label"].cat.categories
            colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(estimator_names) < 10 else len(estimator_names),
            )
            for report_idx, estimator_name in enumerate(estimator_names):
                query = f"estimator_name == '{estimator_name}'"
                roc_curve = self.roc_curve.query(query)
                roc_auc = self.roc_auc.query(query)["roc_auc"]

                line_kwargs["color"] = colors[report_idx]
                line_kwargs["alpha"] = 0.6
                line_kwargs_validated = _validate_style_kwargs(
                    line_kwargs, roc_curve_kwargs[report_idx]
                )

                for split_idx, segment in roc_curve.groupby(
                    "split_index", observed=True
                ):
                    if split_idx == 0:
                        label_kwargs = {
                            "label": (
                                f"{estimator_name} "
                                f"(AUC = {roc_auc.mean():0.2f} +/- "
                                f"{roc_auc.std():0.2f})"
                            )
                        }
                    else:
                        label_kwargs = {}

                    (line,) = self.ax_.plot(
                        segment["fpr"],
                        segment["tpr"],
                        **(line_kwargs_validated | label_kwargs),
                    )
                    lines.append(line)

            info_pos_label = (
                f"\n(Positive label: {self.pos_label})"
                if self.pos_label is not None
                else ""
            )

            if plot_chance_level:
                self.chance_level_ = _add_chance_level(
                    self.ax_,
                    chance_level_kwargs,
                    self._default_chance_level_kwargs,
                )
            else:
                self.chance_level_ = None

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = "External set"

            _, labels = self.ax_.get_legend_handles_labels()
            if len(labels) > MAX_N_LABELS:  # too many lines to fit legend in the plot
                self.ax_.legend(bbox_to_anchor=(1.02, 1), title=legend_title)
            else:
                self.ax_.legend(loc="lower right", title=legend_title)
            self.ax_.set_title("ROC Curve")

        else:  # multiclass-classification
            info_pos_label = None  # irrelevant for multiclass
            labels = self.roc_curve["label"].cat.categories
            colors = sample_mpl_colormap(
                colormaps.get_cmap("tab10"),
                10 if len(estimator_names) < 10 else len(estimator_names),
            )
            curve_idx = 0

            for est_idx, estimator_name in enumerate(estimator_names):
                est_color = colors[est_idx]

                for label_idx, label in enumerate(labels):
                    query = f"label == {label} & estimator_name == '{estimator_name}'"
                    roc_curve = self.roc_curve.query(query)

                    roc_auc = self.roc_auc.query(query)["roc_auc"]

                    for split_idx, segment in roc_curve.groupby(
                        "split_index", observed=True
                    ):
                        if split_idx == 0:
                            label_kwargs = {
                                "label": (
                                    f"{estimator_name} "
                                    f"(AUC = {roc_auc.mean():0.2f} +/- "
                                    f"{roc_auc.std():0.2f})"
                                )
                            }
                        else:
                            label_kwargs = {}

                        line_kwargs["color"] = est_color
                        line_kwargs["alpha"] = 0.6
                        line_kwargs_validated = _validate_style_kwargs(
                            line_kwargs, roc_curve_kwargs[curve_idx]
                        )

                        (line,) = self.ax_[label_idx].plot(
                            segment["fpr"],
                            segment["tpr"],
                            **(line_kwargs_validated | label_kwargs),
                        )
                        lines.append(line)

                        curve_idx += 1

                    info_pos_label = f"\n(Positive label: {label})"
                    _set_axis_labels(self.ax_[label_idx], info_pos_label)

            if plot_chance_level:
                self.chance_level_ = []
                for ax in self.ax_:
                    self.chance_level_.append(
                        _add_chance_level(
                            ax,
                            chance_level_kwargs,
                            self._default_chance_level_kwargs,
                        )
                    )
            else:
                self.chance_level_ = None

            if self.data_source in ("train", "test"):
                legend_title = f"{self.data_source.capitalize()} set"
            else:
                legend_title = "External set"

            for ax in self.ax_:
                _, labels = ax.get_legend_handles_labels()
                ax.legend(loc="lower right", title=legend_title)

            self.figure_.suptitle("ROC Curve")

        return self.ax_, lines, info_pos_label

    @StyleDisplayMixin.style_plot
    def _plot_matplotlib(
        self,
        *,
        estimator_name: str | None = None,
        roc_curve_kwargs: dict[str, Any] | list[dict[str, Any]] | None = None,
        plot_chance_level: bool = True,
        chance_level_kwargs: dict[str, Any] | None = None,
        despine: bool = True,
    ) -> None:
        """Plot visualization.

        Extra keyword arguments will be passed to matplotlib's ``plot``.

        Parameters
        ----------
        estimator_name : str, default=None
            Name of the estimator used to plot the ROC curve. If `None`, we use
            the inferred name from the estimator.

        roc_curve_kwargs : dict or list of dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the ROC curve(s).

        plot_chance_level : bool, default=True
            Whether to plot the chance level.

        chance_level_kwargs : dict, default=None
            Keyword arguments to be passed to matplotlib's `plot` for rendering
            the chance level line.

        despine : bool, default=True
            Whether to remove the top and right spines from the plot.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> display = report.metrics.roc()
        >>> display.plot(roc_curve_kwargs={"color": "tab:red"})
        """
        if (
            self.report_type == "comparison-cross-validation"
            and self.ml_task == "multiclass-classification"
        ):
            n_labels = len(self.roc_auc["label"].cat.categories)
            self.figure_, self.ax_ = plt.subplots(
                ncols=n_labels, figsize=(6.4 * n_labels, 4.8)
            )
        else:
            self.figure_, self.ax_ = plt.subplots()

        if roc_curve_kwargs is None:
            roc_curve_kwargs = self._default_roc_curve_kwargs

        if self.ml_task == "binary-classification":
            n_curves = len(self.roc_auc.query(f"label == {self.pos_label!r}"))
        else:
            n_curves = len(self.roc_auc)

        roc_curve_kwargs = self._validate_curve_kwargs(
            curve_param_name="roc_curve_kwargs",
            curve_kwargs=roc_curve_kwargs,
            n_curves=n_curves,
            report_type=self.report_type,
        )

        if self.report_type == "estimator":
            self.ax_, self.lines_, info_pos_label = self._plot_single_estimator(
                estimator_name=(
                    estimator_name
                    or self.roc_auc["estimator_name"].cat.categories.item()
                ),
                roc_curve_kwargs=roc_curve_kwargs,
                plot_chance_level=plot_chance_level,
                chance_level_kwargs=chance_level_kwargs,
            )
        elif self.report_type == "cross-validation":
            self.ax_, self.lines_, info_pos_label = (
                self._plot_cross_validated_estimator(
                    estimator_name=(
                        estimator_name
                        or self.roc_auc["estimator_name"].cat.categories.item()
                    ),
                    roc_curve_kwargs=roc_curve_kwargs,
                    plot_chance_level=plot_chance_level,
                    chance_level_kwargs=chance_level_kwargs,
                )
            )
        elif self.report_type == "comparison-estimator":
            self.ax_, self.lines_, info_pos_label = self._plot_comparison_estimator(
                estimator_names=self.roc_auc["estimator_name"].cat.categories,
                roc_curve_kwargs=roc_curve_kwargs,
                plot_chance_level=plot_chance_level,
                chance_level_kwargs=chance_level_kwargs,
            )
        elif self.report_type == "comparison-cross-validation":
            self.ax_, self.lines_, info_pos_label = (
                self._plot_comparison_cross_validation(
                    estimator_names=self.roc_auc["estimator_name"].cat.categories,
                    roc_curve_kwargs=roc_curve_kwargs,
                    plot_chance_level=plot_chance_level,
                    chance_level_kwargs=chance_level_kwargs,
                )
            )
        else:
            raise ValueError(
                "`report_type` should be one of 'estimator', 'cross-validation', "
                "'comparison-cross-validation' or 'comparison-estimator'. "
                f"Got '{self.report_type}' instead."
            )

        if (
            self.report_type == "comparison-cross-validation"
            and self.ml_task == "multiclass-classification"
        ):
            for ax in self.ax_:
                if despine:
                    _despine_matplotlib_axis(ax)
        else:
            _set_axis_labels(self.ax_, info_pos_label)

            if despine:
                _despine_matplotlib_axis(self.ax_)

    @classmethod
    def _compute_data_for_display(
        cls,
        y_true: Sequence[YPlotData],
        y_pred: Sequence[YPlotData],
        *,
        report_type: ReportType,
        estimators: Sequence[BaseEstimator],
        ml_task: MLTask,
        data_source: Literal["train", "test", "X_y"],
        pos_label: PositiveLabel | None,
        drop_intermediate: bool = True,
    ) -> "RocCurveDisplay":
        """Private method to create a RocCurveDisplay from predictions.

        Parameters
        ----------
        y_true : list of array-like of shape (n_samples,)
            True binary labels in binary classification.

        y_pred : list of ndarray of shape (n_samples,)
            Target scores, can either be probability estimates of the positive class,
            confidence values, or non-thresholded measure of decisions (as returned by
            "decision_function" on some classifiers).

        report_type : {"comparison-cross-validation", "comparison-estimator", \
                "cross-validation", "estimator"}
            The type of report.

        estimators : list of estimator instances
            The estimators from which `y_pred` is obtained.

        ml_task : {"binary-classification", "multiclass-classification"}
            The machine learning task.

        data_source : {"train", "test", "X_y"}
            The data source used to compute the ROC curve.

        pos_label : int, float, bool or str, default=None
            The class considered as the positive class when computing the
            precision and recall metrics.

        drop_intermediate : bool, default=True
            Whether to drop intermediate points with identical value.

        Returns
        -------
        display : RocCurveDisplay
            Object that stores computed values.
        """
        pos_label_validated = cls._validate_from_predictions_params(
            y_true, y_pred, ml_task=ml_task, pos_label=pos_label
        )

        roc_curve_records = []
        roc_auc_records = []

        if ml_task == "binary-classification":
            for y_true_i, y_pred_i in zip(y_true, y_pred, strict=False):
                fpr_i, tpr_i, thresholds_i = roc_curve(
                    y_true_i.y,
                    y_pred_i.y,
                    pos_label=pos_label,
                    drop_intermediate=drop_intermediate,
                )
                roc_auc_i = auc(fpr_i, tpr_i)

                pos_label_validated = cast(PositiveLabel, pos_label_validated)

                for fpr, tpr, threshold in zip(
                    fpr_i, tpr_i, thresholds_i, strict=False
                ):
                    roc_curve_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "label": pos_label_validated,
                            "threshold": threshold,
                            "fpr": fpr,
                            "tpr": tpr,
                        }
                    )

                roc_auc_records.append(
                    {
                        "estimator_name": y_true_i.estimator_name,
                        "split_index": y_true_i.split_index,
                        "label": pos_label_validated,
                        "roc_auc": roc_auc_i,
                    }
                )

        else:  # multiclass-classification
            # OvR fashion to collect fpr, tpr, and roc_auc
            for y_true_i, y_pred_i, est in zip(
                y_true, y_pred, estimators, strict=False
            ):
                label_binarizer = LabelBinarizer().fit(est.classes_)
                y_true_onehot_i: NDArray = label_binarizer.transform(y_true_i.y)
                for class_idx, class_ in enumerate(est.classes_):
                    fpr_class_i, tpr_class_i, thresholds_class_i = roc_curve(
                        y_true_onehot_i[:, class_idx],
                        y_pred_i.y[:, class_idx],
                        pos_label=None,
                        drop_intermediate=drop_intermediate,
                    )
                    roc_auc_class_i = auc(fpr_class_i, tpr_class_i)

                    for fpr, tpr, threshold in zip(
                        fpr_class_i, tpr_class_i, thresholds_class_i, strict=False
                    ):
                        roc_curve_records.append(
                            {
                                "estimator_name": y_true_i.estimator_name,
                                "split_index": y_true_i.split_index,
                                "label": class_,
                                "threshold": threshold,
                                "fpr": fpr,
                                "tpr": tpr,
                            }
                        )

                    roc_auc_records.append(
                        {
                            "estimator_name": y_true_i.estimator_name,
                            "split_index": y_true_i.split_index,
                            "label": class_,
                            "roc_auc": roc_auc_class_i,
                        }
                    )

        dtypes = {
            "estimator_name": "category",
            "split_index": "category",
            "label": "category",
        }

        return cls(
            roc_curve=DataFrame.from_records(roc_curve_records).astype(dtypes),
            roc_auc=DataFrame.from_records(roc_auc_records).astype(dtypes),
            pos_label=pos_label_validated,
            data_source=data_source,
            ml_task=ml_task,
            report_type=report_type,
        )

    def frame(self, with_roc_auc: bool = False) -> DataFrame:
        """Get the data used to create the ROC curve plot.

        Parameters
        ----------
        with_roc_auc : bool, default=False
            Whether to include ROC AUC scores in the output DataFrame.

        Returns
        -------
        DataFrame
            A DataFrame containing the ROC curve data with columns depending on the
            report type:

            - `estimator_name`: Name of the estimator (when comparing estimators)
            - `split_index`: Cross-validation fold ID (when doing cross-validation)
            - `label`: Class label (for multiclass-classification)
            - `threshold`: Decision threshold
            - `fpr`: False Positive Rate
            - `tpr`: True Positive Rate
            - `roc_auc`: Area Under the Curve (when `with_roc_auc=True`)

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import EstimatorReport, train_test_split
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> clf = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(clf, **split_data)
        >>> display = report.metrics.roc()
        >>> df = display.frame()
        """
        if with_roc_auc:  # noqa: SIM108
            # The merge between the ROC curve and the ROC AUC is done without
            # specifying the columns to merge on, hence done on all columns that are
            # present in both DataFrames.
            # In this case, the common columns are all columns excepts the ones
            # containing the statistics.
            df = self.roc_curve.merge(self.roc_auc)
        else:
            df = self.roc_curve

        statistical_columns = ["threshold", "fpr", "tpr"]
        if with_roc_auc:
            statistical_columns.append("roc_auc")

        if self.report_type == "estimator":
            indexing_columns = []
        elif self.report_type == "cross-validation":
            indexing_columns = ["split_index"]
        elif self.report_type == "comparison-estimator":
            indexing_columns = ["estimator_name"]
        else:  # self.report_type == "comparison-cross-validation"
            indexing_columns = ["estimator_name", "split_index"]

        if self.ml_task == "binary-classification":
            columns = indexing_columns + statistical_columns
        else:
            columns = indexing_columns + ["label"] + statistical_columns

        return df[columns]



from __future__ import annotations

from collections.abc import Callable, Iterable
from typing import Any, Literal, cast

import joblib
import numpy as np
import pandas as pd
from numpy.typing import ArrayLike
from sklearn import metrics
from sklearn.base import is_classifier, is_regressor
from sklearn.inspection import permutation_importance
from sklearn.metrics import make_scorer
from sklearn.pipeline import Pipeline
from sklearn.utils.metaestimators import available_if

from skore._externals._pandas_accessors import DirNamesMixin
from skore._sklearn._base import _BaseAccessor
from skore._sklearn._estimator.report import EstimatorReport
from skore._sklearn.types import Aggregate
from skore._utils._accessor import _check_has_coef, _check_has_feature_importances
from skore._utils._index import flatten_multi_index

DataSource = Literal["test", "train", "X_y"]

Metric = Literal[
    "accuracy",
    "precision",
    "recall",
    "brier_score",
    "roc_auc",
    "log_loss",
    "r2",
    "rmse",
]

# If scoring represents a single score, one can use:
#   - a single string (see The scoring parameter: defining model evaluation rules);
#   - a callable (see Callable scorers) that returns a single value.
# If scoring represents multiple scores, one can use:
#   - a list or tuple of unique strings;
#   - a callable returning a dictionary where the keys are the metric names
#   and the values are the metric scores;
#   - a dictionary with metric names as keys and callables a values.
Scoring = Metric | Callable | Iterable[Metric] | dict[str, Callable]

metric_to_scorer: dict[Metric, Callable] = {
    "accuracy": make_scorer(metrics.accuracy_score),
    "precision": make_scorer(metrics.precision_score),
    "recall": make_scorer(metrics.recall_score),
    "brier_score": make_scorer(metrics.brier_score_loss),
    "roc_auc": make_scorer(metrics.roc_auc_score),
    "log_loss": make_scorer(metrics.log_loss),
    "r2": make_scorer(metrics.r2_score),
    "rmse": make_scorer(metrics.root_mean_squared_error),
}


def _check_scoring(scoring: Any) -> Scoring | None:
    """Check that `scoring` is valid, and convert it to a suitable form as needed.

    If `scoring` is a list of strings, it is checked against our own metric names.
    For example, "rmse" is recognized as root-mean-square error, even though sklearn
    itself does not recognize this name.
    Similarly, "neg_root_mean_square_error" is not recognized, and leads to an error.

    Parameters
    ----------
    scoring : str, callable, list, tuple, dict, or None
        The scoring to check.

    Returns
    -------
    scoring
        A scoring hopefully suitable for passing to `permutation_importance`.
        Can be equal to the original scoring.

    Raises
    ------
    TypeError
        If `scoring` does not type-check.

    Examples
    --------
    >>> from sklearn.metrics import make_scorer, root_mean_squared_error
    >>> from skore._sklearn._estimator.feature_importance_accessor import _check_scoring

    >>> _check_scoring(None)  # Returns None

    >>> _check_scoring(make_scorer(root_mean_squared_error))
    make_scorer(root_mean_squared_error, ...)

    >>> _check_scoring({"rmse": make_scorer(root_mean_squared_error)})
    {'rmse': make_scorer(root_mean_squared_error, ...)}

    >>> _check_scoring("rmse")
    {'rmse': make_scorer(root_mean_squared_error, ...)}

    >>> _check_scoring(["r2", "rmse"])
    {'r2': make_scorer(r2_score, ...),
    'rmse': make_scorer(root_mean_squared_error, ...)}

    >>> _check_scoring("neg_root_mean_squared_error")
    Traceback (most recent call last):
    TypeError: If scoring is a string, it must be one of ...;
    got 'neg_root_mean_squared_error'

    >>> _check_scoring(["r2", make_scorer(root_mean_squared_error)])
    Traceback (most recent call last):
    TypeError: If scoring is a list or tuple, it must contain only strings; ...

    >>> _check_scoring(3)
    Traceback (most recent call last):
    TypeError: scoring must be a string, callable, list, tuple or dict;
    got <class 'int'>
    """
    if scoring is None or callable(scoring) or isinstance(scoring, dict):
        return scoring
    elif isinstance(scoring, str):
        if scoring in metric_to_scorer:
            # Convert to scorer
            return {scoring: metric_to_scorer[cast(Metric, scoring)]}
        raise TypeError(
            "If scoring is a string, it must be one of "
            f"{list(metric_to_scorer.keys())}; got '{scoring}'"
        )
    elif isinstance(scoring, list | tuple):
        result: dict[str, Callable] = {}
        for s in scoring:
            if isinstance(s, str):
                result |= cast(dict[str, Callable], _check_scoring(s))
            else:
                raise TypeError(
                    "If scoring is a list or tuple, it must contain only strings; "
                    f"got {s} of type {type(s)}"
                )
        return result
    else:
        raise TypeError(
            "scoring must be a string, callable, list, tuple or dict; "
            f"got {type(scoring)}"
        )


class _FeatureImportanceAccessor(_BaseAccessor[EstimatorReport], DirNamesMixin):
    """Accessor for feature importance related operations.

    You can access this accessor using the `feature_importance` attribute.
    """

    def __init__(self, parent: EstimatorReport) -> None:
        super().__init__(parent)

    @available_if(_check_has_coef())
    def coefficients(self) -> pd.DataFrame:
        """Retrieve the coefficients of a linear model, including the intercept.

        Examples
        --------
        >>> from sklearn.datasets import load_diabetes
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_diabetes(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> regressor = Ridge()
        >>> report = EstimatorReport(regressor, **split_data)
        >>> report.feature_importance.coefficients()
                   Coefficient
        Intercept     152.4...
        Feature #0     21.2...
        Feature #1    -60.4...
        Feature #2    302.8...
        Feature #3    179.4...
        Feature #4      8.9...
        Feature #5    -28.8...
        Feature #6   -149.3...
        Feature #7    112.6...
        Feature #8    250.5...
        Feature #9     99.5...
        """
        parent_estimator = self._parent.estimator_

        if isinstance(parent_estimator, Pipeline):
            feature_names = parent_estimator[:-1].get_feature_names_out()
        else:
            if hasattr(parent_estimator, "feature_names_in_"):
                feature_names = parent_estimator.feature_names_in_
            else:
                feature_names = [
                    f"Feature #{i}" for i in range(parent_estimator.n_features_in_)
                ]

        estimator = (
            parent_estimator[-1]
            if isinstance(parent_estimator, Pipeline)
            else parent_estimator
        )
        try:
            intercept = np.atleast_2d(estimator.intercept_)
        except AttributeError:
            # TransformedTargetRegressor() does not expose `intercept_`
            intercept = np.atleast_2d(estimator.regressor_.intercept_)
            # Uncomment when SGDOneClassSVM is fully supported by EstimatorReport
            # except AttributeError:
            # SGDOneClassSVM does not expose `intercept_`
            # intercept = None

        try:
            coef = np.atleast_2d(estimator.coef_)
        except AttributeError:
            # TransformedTargetRegressor() does not expose `coef_`
            coef = np.atleast_2d(estimator.regressor_.coef_)

        if intercept is None:
            data = coef.T
            index = list(feature_names)
        else:
            data = np.concatenate([intercept, coef.T])
            index = ["Intercept"] + list(feature_names)

        if data.shape[1] == 1:
            columns = ["Coefficient"]
        elif is_classifier(parent_estimator):
            columns = [f"Class #{i}" for i in range(data.shape[1])]
        else:
            columns = [f"Target #{i}" for i in range(data.shape[1])]

        df = pd.DataFrame(
            data=data,
            index=index,
            columns=columns,
        )

        return df

    @available_if(_check_has_feature_importances())
    def mean_decrease_impurity(self):
        """Retrieve the mean decrease impurity (MDI) of a tree-based model.

        This method is available for estimators that expose a `feature_importances_`
        attribute. See for example
        :attr:`sklearn.ensemble.GradientBoostingClassifier.feature_importances_`.
        In particular, note that the MDI is computed at fit time, i.e. using the
        training data.

        Examples
        --------
        >>> from sklearn.datasets import make_classification
        >>> from sklearn.ensemble import RandomForestClassifier
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = make_classification(n_features=5, random_state=42)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> forest = RandomForestClassifier(n_estimators=5, random_state=0)
        >>> report = EstimatorReport(forest, **split_data)
        >>> report.feature_importance.mean_decrease_impurity()
                   Mean decrease impurity
        Feature #0                0.06...
        Feature #1                0.19...
        Feature #2                0.01...
        Feature #3                0.69...
        Feature #4                0.02...
        """
        parent_estimator = self._parent.estimator_
        estimator = (
            parent_estimator.steps[-1][1]
            if isinstance(parent_estimator, Pipeline)
            else parent_estimator
        )

        data = estimator.feature_importances_

        if isinstance(parent_estimator, Pipeline):
            feature_names = parent_estimator[:-1].get_feature_names_out()
        else:
            if hasattr(parent_estimator, "feature_names_in_"):
                feature_names = parent_estimator.feature_names_in_
            else:
                feature_names = [
                    f"Feature #{i}" for i in range(parent_estimator.n_features_in_)
                ]

        df = pd.DataFrame(
            data=data,
            index=feature_names,
            columns=["Mean decrease impurity"],
        )

        return df

    def permutation(
        self,
        *,
        data_source: DataSource = "test",
        X: ArrayLike | None = None,
        y: ArrayLike | None = None,
        aggregate: Aggregate | None = None,
        scoring: Scoring | None = None,
        n_repeats: int = 5,
        max_samples: float = 1.0,
        n_jobs: int | None = None,
        seed: int | None = None,
        flat_index: bool = False,
    ) -> pd.DataFrame:
        """Report the permutation feature importance.

        This computes the permutation importance using sklearn's
        :func:`~sklearn.inspection.permutation_importance` function,
        which consists in permuting the values of one feature and comparing
        the value of `scoring` between with and without the permutation, which gives an
        indication on the impact of the feature.

        By default, `seed` is set to `None`, which means the function will
        return a different result at every call. In that case, the results are not
        cached. If you wish to take advantage of skore's caching capabilities, make
        sure you set the `seed` parameter.

        Parameters
        ----------
        data_source : {"test", "train", "X_y"}, default="test"
            The data source to use.

            - "test" : use the test set provided when creating the report.
            - "train" : use the train set provided when creating the report.
            - "X_y" : use the provided `X` and `y` to compute the metric.

        X : array-like of shape (n_samples, n_features), default=None
            New data on which to compute the metric. By default, we use the test
            set provided when creating the report.

        y : array-like of shape (n_samples,), default=None
            New target on which to compute the metric. By default, we use the test
            target provided when creating the report.

        aggregate : {"mean", "std"} or list of such str, default=None
            Function to aggregate the scores across the repeats.

        scoring : str, callable, list, tuple, or dict, default=None
            The scorer to pass to :func:`~sklearn.inspection.permutation_importance`.

            If `scoring` represents a single score, one can use:

            - a single string, which must be one of the supported metrics;
            - a callable that returns a single value.

            If `scoring` represents multiple scores, one can use:

            - a list or tuple of unique strings, which must be one of the supported
              metrics;
            - a callable returning a dictionary where the keys are the metric names
              and the values are the metric scores;
            - a dictionary with metric names as keys and callables a values.

        n_repeats : int, default=5
            Number of times to permute a feature.

        max_samples : int or float, default=1.0
            The number of samples to draw from `X` to compute feature importance
            in each repeat (without replacement).

            - If int, then draw max_samples samples.
            - If float, then draw max_samples * X.shape[0] samples.
            - If max_samples is equal to 1.0 or X.shape[0], all samples will be used.

            While using this option may provide less accurate importance estimates,
            it keeps the method tractable when evaluating feature importance on
            large datasets. In combination with n_repeats, this allows to control
            the computational speed vs statistical accuracy trade-off of this method.

        n_jobs : int or None, default=None
            Number of jobs to run in parallel. -1 means using all processors.

        seed : int or None, default=None
            The seed used to initialize the random number generator used for the
            permutations.

        flat_index : bool, default=False
            Whether to flatten the multi-index columns. Flat index will always be lower
            case, do not include spaces and remove the hash symbol to ease indexing.

        Returns
        -------
        pandas.DataFrame
            The permutation importance.

        Examples
        --------
        >>> from sklearn.datasets import make_regression
        >>> from sklearn.linear_model import Ridge
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = make_regression(n_features=3, random_state=0)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> regressor = Ridge()
        >>> report = EstimatorReport(regressor, **split_data)

        >>> report.feature_importance.permutation(
        ...    n_repeats=2,
        ...    seed=0,
        ... )
        Repeat              Repeat #0  Repeat #1
        Metric  Feature
        r2      Feature #0   0.699...   0.885...
                Feature #1   2.320...   2.636...
                Feature #2   0.028...   0.022...

        >>> report.feature_importance.permutation(
        ...    scoring=["r2", "rmse"],
        ...    n_repeats=2,
        ...    seed=0,
        ... )
        Repeat             Repeat #0  Repeat #1
        Metric Feature
        r2     Feature #0   0.699...   0.885...
               Feature #1   2.320...   2.636...
               Feature #2   0.028...   0.022...
        rmse   Feature #0 -47.222... -53.231...
               Feature #1 -86.608... -92.366...
               Feature #2  -8.930...  -7.916...

        >>> report.feature_importance.permutation(
        ...    n_repeats=2,
        ...    aggregate=["mean", "std"],
        ...    seed=0,
        ... )
                                mean       std
        Metric  Feature
        r2      Feature #0  0.792...  0.131...
                Feature #1  2.478...  0.223...
                Feature #2  0.025...  0.003...

        >>> report.feature_importance.permutation(
        ...    n_repeats=2,
        ...    aggregate=["mean", "std"],
        ...    flat_index=True,
        ...    seed=0,
        ... )
                          mean       std
        r2_feature_0  0.792...  0.131...
        r2_feature_1  2.478...  0.223...
        r2_feature_2  0.025...  0.003...
        """
        return self._feature_permutation(
            data_source=data_source,
            data_source_hash=None,
            X=X,
            y=y,
            aggregate=aggregate,
            scoring=scoring,
            n_repeats=n_repeats,
            max_samples=max_samples,
            n_jobs=n_jobs,
            seed=seed,
            flat_index=flat_index,
        )

    def _feature_permutation(
        self,
        *,
        data_source: DataSource = "test",
        data_source_hash: int | None = None,
        X: ArrayLike | None = None,
        y: ArrayLike | None = None,
        aggregate: Aggregate | None = None,
        scoring: Scoring | None = None,
        n_repeats: int = 5,
        max_samples: float = 1.0,
        n_jobs: int | None = None,
        seed: int | None = None,
        flat_index: bool = False,
    ) -> pd.DataFrame:
        """Private interface of `feature_permutation` to pass `data_source_hash`.

        `data_source_hash` is either an `int` when we already computed the hash
        and are able to pass it around, or `None` and thus trigger its computation
        in the underlying process.
        """
        checked_scoring = _check_scoring(scoring)

        if data_source_hash is None:
            X_, y_true, data_source_hash = self._get_X_y_and_data_source_hash(
                data_source=data_source, X=X, y=y
            )

        # If seed is None, then we do not do any caching
        if seed is None:
            cache_key = None

        elif isinstance(seed, int):
            # build the cache key components to finally create a tuple that will be used
            # to check if the metric has already been computed
            cache_key_parts: list[Any] = [
                self._parent._hash,
                "permutation_importance",
                data_source,
            ]

            if data_source_hash is not None:
                cache_key_parts.append(data_source_hash)

            if callable(scoring) or isinstance(scoring, list | dict):
                cache_key_parts.append(joblib.hash(scoring))
            else:
                cache_key_parts.append(scoring)

            # aggregate is not included in the cache
            # in order to trade off computation for storage

            # order arguments by key to ensure cache works
            # n_jobs variable should not be in the cache
            kwargs = {
                "n_repeats": n_repeats,
                "max_samples": max_samples,
                "seed": seed,
            }
            for _, v in sorted(kwargs.items()):
                cache_key_parts.append(v)

            cache_key = tuple(cache_key_parts)

        else:
            raise ValueError(f"seed must be an integer or None; got {type(seed)}")

        if cache_key in self._parent._cache:
            score = self._parent._cache[cache_key]
        else:
            sklearn_score = permutation_importance(
                estimator=self._parent.estimator_,
                X=X_,
                y=y_true,
                scoring=checked_scoring,
                n_repeats=n_repeats,
                n_jobs=n_jobs,
                random_state=seed,
                max_samples=max_samples,
            )
            score = sklearn_score.get("importances")

            feature_names = (
                self._parent.estimator_.feature_names_in_
                if hasattr(self._parent.estimator_, "feature_names_in_")
                else [f"Feature #{i}" for i in range(X_.shape[1])]
            )

            # If there is more than one metric
            if score is None:
                data = np.concatenate(
                    [v["importances"] for v in sklearn_score.values()]
                )
                n_repeats = data.shape[1]
                index = pd.MultiIndex.from_product(
                    [sklearn_score, feature_names], names=("Metric", "Feature")
                )
                columns = pd.Index(
                    (f"Repeat #{i}" for i in range(n_repeats)), name="Repeat"
                )
                score = pd.DataFrame(data=data, index=index, columns=columns)
            else:
                data = score
                n_repeats = data.shape[1]

                # Get score name
                if scoring is None:
                    if is_classifier(self._parent.estimator_):
                        scoring_name = "accuracy"
                    elif is_regressor(self._parent.estimator_):
                        scoring_name = "r2"
                else:
                    # e.g. if scoring is a callable
                    scoring_name = None

                    # no other cases to deal with explicitly, because
                    # scoring cannot possibly be a string at this stage

                if scoring_name is None:
                    index = pd.Index(feature_names, name="Feature")
                else:
                    index = pd.MultiIndex.from_product(
                        [[scoring_name], feature_names], names=("Metric", "Feature")
                    )

                columns = pd.Index(
                    (f"Repeat #{i}" for i in range(n_repeats)), name="Repeat"
                )
                score = pd.DataFrame(data=data, index=index, columns=columns)

            if cache_key is not None:
                # Unless seed is an int (i.e. the call is deterministic),
                # we do not cache
                self._parent._cache[cache_key] = score

        if aggregate:
            if isinstance(aggregate, str):
                aggregate = [aggregate]
            score = score.aggregate(func=aggregate, axis=1)

        if flat_index and isinstance(score.index, pd.MultiIndex):
            score.index = flatten_multi_index(score.index)

        return score

    ####################################################################################
    # Methods related to the help tree
    ####################################################################################

    def _format_method_name(self, name: str) -> str:
        return f"{name}(...)".ljust(29)

    def _get_help_panel_title(self) -> str:
        return "[bold cyan]Available feature importance methods[/bold cyan]"

    def _get_help_tree_title(self) -> str:
        return "[bold cyan]report.feature_importance[/bold cyan]"

    def __repr__(self) -> str:
        """Return a string representation using rich."""
        return self._rich_repr(class_name="skore.EstimatorReport.feature_importance")



import os
import sys
from typing import Any


def get_environment_info() -> dict[str, Any]:
    """Detect the current Python execution environment.

    Returns a dictionary with information about the environment.
    """
    env_info: dict[str, Any] = {
        "is_jupyter": False,
        "is_vscode": False,
        "is_interactive": False,
        "environment_name": "standard_python",
        "details": {},
    }

    env_info["is_interactive"] = hasattr(sys, "ps1")

    try:
        # get_ipython() is defined when running in Jupyter or IPython
        # there is no need to import IPython here
        shell = get_ipython().__class__.__name__  # type: ignore
        env_info["details"]["ipython_shell"] = shell

        if shell == "ZMQInteractiveShell":  # Jupyter notebook/lab
            env_info["is_jupyter"] = True
            env_info["environment_name"] = "jupyter"
        elif shell == "TerminalInteractiveShell":  # IPython terminal
            env_info["environment_name"] = "ipython_terminal"
    except NameError:
        pass

    if "VSCODE_PID" in os.environ:
        env_info["is_vscode"] = True
        if env_info["is_interactive"]:
            env_info["environment_name"] = "vscode_interactive"
        else:
            env_info["environment_name"] = "vscode_script"

    env_info["details"]["python_executable"] = sys.executable
    env_info["details"]["python_version"] = sys.version

    return env_info


def is_environment_notebook_like() -> bool:
    """Return `True` if the execution context dcan render HTML. `False` otherwise."""
    info = get_environment_info()
    return info["is_vscode"] or info["is_jupyter"]



"""'Random state is unset' warning.

This warning is shown when `random_state` is unset.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from skore._sklearn.train_test_split.warning.train_test_split_warning import (
    TrainTestSplitWarning,
)

if TYPE_CHECKING:
    from numpy.random import RandomState


class RandomStateUnsetWarning(TrainTestSplitWarning):
    """Check whether `random_state` is set."""

    MSG = (
        "We recommend setting the parameter `random_state`. "
        "This will ensure the reproducibility of your work."
    )

    @staticmethod
    def check(
        shuffle: bool,
        random_state: int | RandomState | None,
        **kwargs,
    ) -> str | None:
        """Check whether ``random_state`` is set.

        Parameters
        ----------
        shuffle : bool
            Whether to shuffle the data before splitting.
            If ``False``, the warning is skipped.
        random_state : int or RandomState instance or None
            Controls the shuffling.
            Pass something other than ``None`` to ensure reproducibility.

        Returns
        -------
        warning
            None if the check passed, otherwise the warning message.
        """
        if shuffle and random_state is None:
            return RandomStateUnsetWarning.MSG
        return None



import json
from typing import Any, Literal

import numpy as np
import pandas as pd
import seaborn as sns
from matplotlib import pyplot as plt
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from skrub import _column_associations
from skrub._reporting._html import to_html
from skrub._reporting._summarize import summarize_dataframe
from skrub._reporting._utils import (
    JSONEncoder,
    duration_to_numeric,
    ellide_string,
    top_k_value_counts,
)

from skore._externals._skrub_compat import sbd
from skore._sklearn._plot.style import StyleDisplayMixin
from skore._sklearn._plot.utils import (
    HelpDisplayMixin,
    PlotBackendMixin,
    _adjust_fig_size,
    _rotate_ticklabels,
    _validate_style_kwargs,
)
from skore._utils._repr_html import ReprHTMLMixin


def _truncate_top_k_categories(
    col: pd.Series | None, k: int, other_label: str = "other"
) -> pd.Series:
    """Truncate a column to the top k most frequent values.

    Replaces the rest with a label defined by ``other_label``. Note that if `col` is not
    a categorical column, it is returned as is.

    Parameters
    ----------
    col : pd.Series or None
        The column to truncate.

    k : int
        The number of most frequent values to keep.

    other_label : str, default="other"
        The label to use for the rest of the values.

    Returns
    -------
    pd.Series
        The truncated column.
    """
    if col is None or sbd.is_numeric(col):
        return col

    col = col.copy()
    _, counter = top_k_value_counts(col, k=k)
    values, _ = zip(*counter, strict=False)
    # we don't want to replace NaN with 'other'
    keep = sbd.is_in(col, values) | sbd.is_null(col)
    if isinstance(col.dtype, pd.CategoricalDtype):
        col = col.cat.add_categories(other_label)
        col[~keep] = other_label
        col = col.cat.remove_unused_categories()
        col = col.cat.rename_categories(
            {v: ellide_string(v, max_len=20) for v in values if isinstance(v, str)}
        )
    else:
        col[~keep] = other_label
        col = col.apply(
            lambda x: ellide_string(x, max_len=20) if isinstance(x, str) else x
        )
    return col


def _compute_contingency_table(
    x: pd.Series, y: pd.Series, hue: pd.Series | None, k: int
) -> pd.DataFrame:
    """Compute the contingency table of x and y with filtering only to the top k pairs.

    The contingency table is a symmetric matrix where the values are the mean of hue
    for a given pair of (x, y) entries. If hue is not provided, the values are the
    frequency of each pair (x, y).

    Parameters
    ----------
    x : pd.Series
        The first column to aggregate.

    y : pd.Series
        The second column to aggregate.

    hue : pd.Series | None
        The column to aggregate by.

    k : int
        The number of top pairs to select.

    Returns
    -------
    pd.DataFrame
        The contingency table.
    """
    if sbd.name(x) is None or sbd.name(y) is None:
        raise ValueError("The series x and y must have a name.")

    if hue is None:
        contingency_table = pd.crosstab(index=y, columns=x)
    else:
        contingency_table = pd.crosstab(index=y, columns=x, values=hue, aggfunc="mean")

    cols = sbd.concat(sbd.to_frame(x), sbd.to_frame(y), axis=1)
    top_pairs = cols.value_counts().nlargest(k).index

    top_x_values = sorted(set(pair[0] for pair in top_pairs))
    top_y_values = sorted(set(pair[1] for pair in top_pairs))

    # As stated by a pandas warning, we call explicitly `infer_objects` to downcast
    # the contingency table and silence the warning using the context manager.
    with pd.option_context("future.no_silent_downcasting", True):
        return (
            contingency_table.fillna(0)
            .infer_objects(copy=False)
            .reindex(index=top_y_values, columns=top_x_values, fill_value=0)
        )


def _resize_categorical_axis(
    *,
    figure: Figure,
    ax: Axes,
    n_categories: int,
    is_x_axis: bool,
    size_per_category: float = 0.3,
) -> None:
    """Resize the axis along which categories are plotted.

    Parameters
    ----------
    fig : Figure
        The figure to resize.

    ax : Axes
        The axis to resize.

    is_x_axis : bool
        Whether the axis is the x-axis.

    size_per_category : float, default=0.3
        The size of the category in the axis. When having annotations, the size needs
        to be tweaked based on the formatting of the annotations.
    """
    if is_x_axis:
        ax.tick_params(axis="x", length=0)
        target_height = figure.get_size_inches()[1]
        target_width = n_categories * size_per_category
    else:
        ax.tick_params(axis="y", length=0)
        target_height = n_categories * size_per_category
        target_width = figure.get_size_inches()[0]
    _adjust_fig_size(figure, ax, target_width, target_height)


class TableReportDisplay(
    StyleDisplayMixin, HelpDisplayMixin, ReprHTMLMixin, PlotBackendMixin
):
    """Display reporting information about a given dataset.

    This display summarizes the dataset and provides a way to visualize
    the distribution of its columns.

    Parameters
    ----------
    summary : dict
        The summary of the dataset, as returned by ``summarize_dataframe``.

    Attributes
    ----------
    ax_ : matplotlib axes
        The axes of the figure.

    figure_ : matplotlib figure.
        The figure of the plot.

    Examples
    --------
    >>> from sklearn.datasets import load_breast_cancer
    >>> from sklearn.linear_model import LogisticRegression
    >>> from skore import train_test_split
    >>> from skore import EstimatorReport
    >>> X, y = load_breast_cancer(return_X_y=True)
    >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
    >>> classifier = LogisticRegression(max_iter=10_000)
    >>> report = EstimatorReport(classifier, **split_data)
    >>> display = report.data.analyze()
    >>> display.plot(kind="corr")
    """

    _default_heatmap_kwargs: dict[str, Any] | None = None
    _default_boxplot_kwargs: dict[str, Any] | None = None
    _default_scatterplot_kwargs: dict[str, Any] | None = None
    _default_stripplot_kwargs: dict[str, Any] | None = None
    _default_histplot_kwargs: dict[str, Any] | None = None

    def __init__(self, summary: dict[str, Any]) -> None:
        self.summary = summary

    @classmethod
    def _compute_data_for_display(cls, dataset: pd.DataFrame) -> "TableReportDisplay":
        """Private method to create a TableReportDisplay from a dataset.

        Parameters
        ----------
        dataset : DataFrame
            The dataset to summarize.

        Returns
        -------
        display : TableReportDisplay
            Object that stores computed values.
        """
        return cls(summarize_dataframe(dataset, with_plots=True, title=None, verbose=0))

    @StyleDisplayMixin.style_plot
    def _plot_matplotlib(
        self,
        *,
        x: str | None = None,
        y: str | None = None,
        hue: str | None = None,
        kind: Literal["dist", "corr"] = "dist",
        top_k_categories: int = 20,
        scatterplot_kwargs: dict[str, Any] | None = None,
        stripplot_kwargs: dict[str, Any] | None = None,
        boxplot_kwargs: dict[str, Any] | None = None,
        heatmap_kwargs: dict[str, Any] | None = None,
        histplot_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Plot distribution or correlation of the columns from the dataset.

        Parameters
        ----------
        x : str, default=None
            The name of the column to use for the x-axis of the plot. Mandatory when
            ``kind='dist'``.

        y : str, default=None
            The name of the column to use for the y-axis of the plot. Only used when
            ``kind='dist'``.

        hue : str, default=None
            The name of the column to use for the color or hue axis of the plot. Only
            used when ``kind='dist'``.

        kind : {'dist', 'corr'}, default='dist'
            The kind of plot drawn.

            - If ``'dist'``, plot a distribution parametrized by ``x``, ``y`` and
              ``hue``. When only ``x`` is defined, the distribution is 1d. When ``y`` is
              also defined, the plot is the 2d. Finally, when the color is set using
              ``hue``, the distribution is 2d, with a color per data-point based on
              ``hue``. This mode handle both numeric and categorical columns.
            - If ``'corr'``, plot `Cramer's V
              <https://en.wikipedia.org/wiki/Cram%C3%A9r%27s_V>`_ correlation among all
              columns. This option doesn't take any ``x``, ``y`` or ``hue`` argument.

        top_k_categories : int, default=20
            For categorical columns, the number of most frequent elements to display.
            Only used when ``kind='dist'``.

        scatterplot_kwargs: dict, default=None
            Keyword arguments to be passed to seaborn's :func:`seaborn.scatterplot` for
            rendering the distribution 2D plot, when both ``x`` and ``y`` are numeric.

        stripplot_kwargs: dict, default=None
            Keyword arguments to be passed to seaborn's :func:`searborn.stripplot` for
            rendering the distribution 2D plot, when either ``x`` or ``y`` is numeric,
            and the other is categorical. This plot is drawn on top of the boxplot.

        boxplot_kwargs: dict, default=None
            Keyword arguments to be passed to seaborn's :func:`seaborn.boxplot` for
            rendering the distribution 2D plot, when either ``x`` or ``y`` is numeric,
            and the other is categorical. This plot is drawn below the stripplot.

        heatmap_kwargs: dict, default=None
            Keyword arguments to be passed to seaborn's :func:`seaborn.heatmap` for
            rendering Cramer's V correlation matrix, when ``kind='corr'`` or when
            ``kind='dist'`` and both ``x`` and ``y`` are categorical.

        histplot_kwargs: dict, default=None
            Keyword arguments to be passed to seaborn's :func:`seaborn.histplot` for
            rendering the distribution 1D plot, when only ``x`` is provided.

        Examples
        --------
        >>> from sklearn.datasets import load_breast_cancer
        >>> from sklearn.linear_model import LogisticRegression
        >>> from skore import train_test_split
        >>> from skore import EstimatorReport
        >>> X, y = load_breast_cancer(return_X_y=True)
        >>> split_data = train_test_split(X=X, y=y, random_state=0, as_dict=True)
        >>> classifier = LogisticRegression(max_iter=10_000)
        >>> report = EstimatorReport(classifier, **split_data)
        >>> display = report.data.analyze()
        >>> display.plot(kind="corr")
        """
        self.figure_, self.ax_ = plt.subplots()
        if kind == "dist":
            match (x is None, y is None, hue is None):
                case (True, True, True) | (True, True, False):
                    raise ValueError(
                        "When kind='dist', at least one of x, y must be provided and "
                        "optionally hue. Got x=None, y=None."
                    )
                case (False, True, True) | (True, False, True):
                    if histplot_kwargs is None:
                        histplot_kwargs = self._default_histplot_kwargs

                    self._plot_distribution_1d(
                        x=x, y=y, k=top_k_categories, histplot_kwargs=histplot_kwargs
                    )
                case _:
                    if scatterplot_kwargs is None:
                        scatterplot_kwargs = self._default_scatterplot_kwargs
                    if stripplot_kwargs is None:
                        stripplot_kwargs = self._default_stripplot_kwargs
                    if boxplot_kwargs is None:
                        boxplot_kwargs = self._default_boxplot_kwargs
                    self._plot_distribution_2d(
                        x=x,
                        y=y,
                        hue=hue,
                        k=top_k_categories,
                        scatterplot_kwargs=scatterplot_kwargs,
                        stripplot_kwargs=stripplot_kwargs,
                        boxplot_kwargs=boxplot_kwargs,
                        heatmap_kwargs=heatmap_kwargs,
                    )

        elif kind == "corr":
            for param_name, param_value in zip(
                ("x", "y", "hue"), (x, y, hue), strict=True
            ):
                if param_value is not None:
                    raise ValueError(
                        f"When {kind=!r}, {param_name!r} argument must be None."
                    )
            if heatmap_kwargs is None:
                heatmap_kwargs = self._default_heatmap_kwargs
            self._plot_cramer(heatmap_kwargs=heatmap_kwargs)

        else:
            raise ValueError(f"'kind' options are 'dist', 'corr', got {kind!r}.")

    def _plot_distribution_1d(
        self,
        *,
        x: str | None,
        y: str | None,
        k: int,
        histplot_kwargs: dict[str, Any] | None,
    ) -> None:
        """Plot 1-dimensional distribution of a feature.

        Parameters
        ----------
        x : str, default=None
            The name of the column to plot.

        y : str, default=None
            The name of the column to plot.

        k : int
            The number of most frequent categories to plot.

        histplot_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's :ref:`histplot
            <https://seaborn.pydata.org/generated/seaborn.histplot.html>`_ for rendering
            the distribution 1D plot.
        """
        default_histplot_kwargs: dict[str, Any] = {}

        column = sbd.col(self.summary["dataframe"], x or y)

        duration_unit = None
        if sbd.is_duration(column):
            column, duration_unit = duration_to_numeric(column)

        if is_categorical := not (sbd.is_numeric(column) or sbd.is_any_date(column)):
            top_k = sbd.col(sbd.head(sbd.value_counts(column), k), "value")

            column = sbd.filter(column, sbd.is_in(column, top_k))
            column = sbd.to_pandas(column)
            column = column.astype(
                pd.CategoricalDtype(categories=sbd.to_list(top_k), ordered=True)
            )
            default_histplot_kwargs["color"] = "tab:orange"

        if sbd.is_integer(column) or is_categorical:
            default_histplot_kwargs["discrete"] = True

        histplot_kwargs_validated = _validate_style_kwargs(
            default_histplot_kwargs, histplot_kwargs or {}
        )

        if x is not None:
            histplot_params = dict(x=column)
            despine_params = dict(bottom=is_categorical)
            if duration_unit is not None:
                self.ax_.set(xlabel=f"{duration_unit.capitalize()}s")
        else:  # y is not None
            histplot_params = dict(y=column)
            despine_params = dict(left=is_categorical)
            if duration_unit is not None:
                self.ax_.set(ylabel=f"{duration_unit.capitalize()}s")

        sns.histplot(ax=self.ax_, **histplot_params, **histplot_kwargs_validated)
        sns.despine(
            self.figure_,
            top=True,
            right=True,
            trim=True,
            offset=10,
            **despine_params,
        )

        if is_categorical:
            _resize_categorical_axis(
                figure=self.figure_,
                ax=self.ax_,
                n_categories=sbd.n_unique(column),
                is_x_axis=x is not None,
            )

        if x is not None and any(
            len(label.get_text()) > 1 for label in self.ax_.get_xticklabels()
        ):
            # rotate only for string longer than 1 character
            _rotate_ticklabels(self.ax_, rotation=45)

    def _plot_distribution_2d(
        self,
        *,
        x: str | None,
        y: str | None,
        hue: str | None = None,
        k: int = 20,
        heatmap_kwargs: dict[str, Any] | None = None,
        stripplot_kwargs: dict[str, Any] | None = None,
        boxplot_kwargs: dict[str, Any] | None = None,
        scatterplot_kwargs: dict[str, Any] | None = None,
    ) -> None:
        """Plot 2-dimensional distribution of two features.

        This function plots a 2-dimensional distribution of two features, with optional
        hue and k-most frequent categories.

        Parameters
        ----------
        x : str or None
            The name of the column to use for the x-axis of the plot.

        y : str or None
            The name of the column to use for the y-axis of the plot.

        hue : str, default=None
            The name of the column to use for the color or hue axis of the plot.

        k : int, default=20
            The number of most frequent categories to plot.

        heatmap_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's heatmap.

        stripplot_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's stripplot.

        boxplot_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's boxplot.

        scatterplot_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's scatterplot.
        """
        x = sbd.col(self.summary["dataframe"], x) if x is not None else None
        y = sbd.col(self.summary["dataframe"], y) if y is not None else None
        hue = sbd.col(self.summary["dataframe"], hue) if hue is not None else None
        x, y = hue if x is None else x, y if y is not None else hue

        despine_params = dict(top=True, right=True, trim=True, offset=10)
        is_x_num, is_y_num = sbd.is_numeric(x), sbd.is_numeric(y)
        hue = _truncate_top_k_categories(hue, k)

        if is_x_num and is_y_num:
            scatterplot_kwargs_validated = _validate_style_kwargs(
                {"alpha": 0.1}, scatterplot_kwargs or {}
            )
            sns.scatterplot(
                x=x,
                y=y,
                hue=hue,
                ax=self.ax_,
                **scatterplot_kwargs_validated,
            )
        elif is_x_num or is_y_num:
            stripplot_kwargs_validated = _validate_style_kwargs(
                {
                    "dodge": False,
                    "size": 6,
                    "alpha": 0.5,
                    "zorder": 0,
                },
                stripplot_kwargs or {},
            )
            boxplot_kwargs_validated = _validate_style_kwargs(
                {
                    "fliersize": 0,
                    "width": 0.5,
                    "whis": (0, 100),  # spellchecker:disable-line
                    "color": "black",
                    "fill": False,
                    "zorder": 1,
                    "boxprops": {"alpha": 0.5},
                    "whiskerprops": {"alpha": 0.5},
                    "capprops": {"alpha": 0.5},
                    "medianprops": {"alpha": 0.5},
                },
                boxplot_kwargs or {},
            )

            if is_x_num:
                y = _truncate_top_k_categories(y, k)
            else:
                x = _truncate_top_k_categories(x, k)

            sns.boxplot(x=x, y=y, ax=self.ax_, **boxplot_kwargs_validated)
            sns.stripplot(x=x, y=y, hue=hue, ax=self.ax_, **stripplot_kwargs_validated)

            _resize_categorical_axis(
                figure=self.figure_,
                ax=self.ax_,
                n_categories=sbd.n_unique(y) if is_x_num else sbd.n_unique(x),
                is_x_axis=not is_x_num,
            )
            if is_x_num:
                despine_params["left"] = True
            else:
                despine_params["bottom"] = True
                if any(
                    len(label.get_text()) > 1 for label in self.ax_.get_xticklabels()
                ):
                    _rotate_ticklabels(self.ax_, rotation=45)
        else:
            if (hue is not None) and (not sbd.is_numeric(hue)):
                raise ValueError(
                    "If 'x' and 'y' are categories, 'hue' must be continuous."
                )

            contingency_table = _compute_contingency_table(x, y, hue, k)
            contingency_table.index = [
                ellide_string(s) for s in contingency_table.index
            ]
            contingency_table.columns = [
                ellide_string(s) for s in contingency_table.columns
            ]

            if max_value := contingency_table.max(axis=None) < 100_000:  # noqa: SIM108
                # avoid scientific notation for small numbers
                annotation_format = (
                    ".0f"
                    if all(
                        pd.api.types.is_integer_dtype(dtype)
                        for dtype in contingency_table.dtypes
                    )
                    else ".2f"
                )
                size_per_category = len(str(max_value)) * 0.15
            else:
                # scientific notation for bigger numbers
                annotation_format = ".2g"
                size_per_category = 0.6

            cbar_kwargs = (
                {"label": f"Mean {sbd.name(hue)}"}
                if hue is not None
                else {"label": "Count"}
            )

            heatmap_kwargs_validated = _validate_style_kwargs(
                {
                    "xticklabels": True,
                    "yticklabels": True,
                    "robust": True,
                    "annot": True,
                    "fmt": annotation_format,
                    "cbar_kws": cbar_kwargs,
                },
                heatmap_kwargs or {},
            )
            sns.heatmap(contingency_table, ax=self.ax_, **heatmap_kwargs_validated)
            despine_params.update(left=True, bottom=True)
            self.ax_.tick_params(axis="both", length=0)

            for is_x_axis, x_or_y in zip(
                [True, False],
                [
                    pd.Series(contingency_table.columns),
                    pd.Series(contingency_table.index),
                ],
                strict=False,
            ):
                _resize_categorical_axis(
                    figure=self.figure_,
                    ax=self.ax_,
                    n_categories=sbd.n_unique(x_or_y),
                    is_x_axis=is_x_axis,
                    size_per_category=size_per_category,
                )

        sns.despine(self.figure_, **despine_params)

        self.ax_.set(xlabel=sbd.name(x), ylabel=sbd.name(y))
        if self.ax_.legend_ is not None:
            sns.move_legend(self.ax_, (1.05, 0.0))

    def _plot_cramer(self, *, heatmap_kwargs: dict[str, Any] | None) -> None:
        """Plot Cramer's V correlation among all columns.

        Parameters
        ----------
        heatmap_kwargs : dict, default=None
            Keyword arguments to be passed to seaborn's heatmap.
        """
        if heatmap_kwargs is None:
            heatmap_kwargs = self._default_heatmap_kwargs or {}

        heatmap_kwargs_validated = _validate_style_kwargs(
            {
                "xticklabels": True,
                "yticklabels": True,
                "robust": True,
                "square": True,
                "annot": True,
                "annot_kws": {"size": 10},
                "fmt": ".2f",
            },
            heatmap_kwargs or {},
        )

        cramer_v_table = pd.DataFrame(
            _column_associations._cramer_v_matrix(self.summary["dataframe"]),
            columns=self.summary["dataframe"].columns,
            index=self.summary["dataframe"].columns,
        )
        # only show the lower triangle of the heatmap since it is a correlation matrix
        # and keep the diagonal as well.
        mask = np.triu(np.ones_like(cramer_v_table, dtype=bool), k=1)

        sns.heatmap(cramer_v_table, mask=mask, ax=self.ax_, **heatmap_kwargs_validated)
        self.ax_.set(title="Cramer's V Correlation")

    def frame(
        self, *, kind: Literal["dataset", "top-associations"] = "dataset"
    ) -> pd.DataFrame:
        """Get the data related to the table report.

        Parameters
        ----------
        kind : {'dataset', 'top-associations'}
            The kind of data to return.

        Returns
        -------
        DataFrame
            The dataset used to create the table report.
        """
        if kind == "dataset":
            return self.summary["dataframe"]
        elif kind == "top-associations":
            return pd.DataFrame(self.summary["top_associations"])
        else:
            raise ValueError(f"Invalid kind: {kind!r}")

    def _html_repr(self) -> str:
        """Show the HTML representation of the report."""
        return to_html(self.summary, standalone=False, column_filters=None)

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__}(...)>"

    def _to_json(self) -> str:
        """Serialize the data of this report in JSON format.

        It is the serialization chosen to be sent to skore-hub.
        """
        to_remove = ["dataframe", "sample_table"]
        data = {k: v for k, v in self.summary.items() if k not in to_remove}
        return json.dumps(data, cls=JSONEncoder)



"""Enhance `sklearn` functions."""

from skore._sklearn._comparison import ComparisonReport
from skore._sklearn._cross_validation import CrossValidationReport
from skore._sklearn._estimator import EstimatorReport
from skore._sklearn._plot import (
    MetricsSummaryDisplay,
    PrecisionRecallCurveDisplay,
    PredictionErrorDisplay,
    RocCurveDisplay,
    TableReportDisplay,
)
from skore._sklearn.find_estimators import find_estimators
from skore._sklearn.train_test_split.train_test_split import train_test_split

__all__ = [
    "ComparisonReport",
    "CrossValidationReport",
    "EstimatorReport",
    "PrecisionRecallCurveDisplay",
    "PredictionErrorDisplay",
    "TableReportDisplay",
    "RocCurveDisplay",
    "MetricsSummaryDisplay",
    "train_test_split",
    "find_estimators",
]



"""Customizations of :mod:`joblib` and :mod:`threadpoolctl` tools for skore usage."""

import functools
import warnings
from functools import update_wrapper

import joblib

from skore._config import config_context, get_config

# Global threadpool controller instance that can be used to locally limit the number of
# threads without looping through all shared libraries every time.
# It should not be accessed directly and _get_threadpool_controller should be used
# instead.
_threadpool_controller = None


def _with_config_and_warning_filters(delayed_func, config, warning_filters):
    """Attach a config to a delayed function."""
    if hasattr(delayed_func, "with_config_and_warning_filters"):
        return delayed_func.with_config_and_warning_filters(config, warning_filters)
    else:
        warnings.warn(
            (
                "`skore._utils._parallel.Parallel` needs to be used in "
                "conjunction with `skore._utils._parallel.delayed` instead of "
                "`joblib.delayed` to correctly propagate the skore configuration to "
                "the joblib workers."
            ),
            UserWarning,
            stacklevel=2,
        )
        return delayed_func


class Parallel(joblib.Parallel):
    """Tweak of :class:`joblib.Parallel` that propagates the skore configuration.

    This subclass of :class:`joblib.Parallel` ensures that the active configuration
    (thread-local) of skore is propagated to the parallel workers for the
    duration of the execution of the parallel tasks.

    The API does not change and you can refer to :class:`joblib.Parallel`
    documentation for more details.
    """

    def __call__(self, iterable):
        """Dispatch the tasks and return the results.

        Parameters
        ----------
        iterable : iterable
            Iterable containing tuples of (delayed_function, args, kwargs) that should
            be consumed.

        Returns
        -------
        results : list
            List of results of the tasks.
        """
        # Capture the thread-local skore configuration at the time
        # Parallel.__call__ is issued since the tasks can be dispatched
        # in a different thread depending on the backend and on the value of
        # pre_dispatch and n_jobs.
        config = get_config()
        warning_filters = warnings.filters
        iterable_with_config_and_warning_filters = (
            (
                _with_config_and_warning_filters(delayed_func, config, warning_filters),
                args,
                kwargs,
            )
            for delayed_func, args, kwargs in iterable
        )
        return super().__call__(iterable_with_config_and_warning_filters)


# remove when https://github.com/joblib/joblib/issues/1071 is fixed
def delayed(function):
    """Capture the arguments of a function to delay its execution.

    This alternative to `joblib.delayed` is meant to be used in conjunction
    with `skore._utils._parallel.Parallel`. The latter captures the skore
    configuration by calling `skore.get_config()` in the current thread, prior to
    dispatching the first task. The captured configuration is then propagated and
    enabled for the duration of the execution of the delayed function in the
    joblib workers.

    Parameters
    ----------
    function : callable
        The function to be delayed.

    Returns
    -------
    output: tuple
        Tuple containing the delayed function, the positional arguments, and the
        keyword arguments.
    """

    @functools.wraps(function)
    def delayed_function(*args, **kwargs):
        return _FuncWrapper(function), args, kwargs

    return delayed_function


class _FuncWrapper:
    """Load the global configuration before calling the function."""

    def __init__(self, function):
        self.function = function
        update_wrapper(self, self.function)

    def with_config_and_warning_filters(self, config, warning_filters):
        self.config = config
        self.warning_filters = warning_filters
        return self

    def __call__(self, *args, **kwargs):
        config = getattr(self, "config", {})
        warning_filters = getattr(self, "warning_filters", [])
        if not config or not warning_filters:
            warnings.warn(
                (
                    "`skore._utils._parallel.delayed` should be used with"
                    " `skore._utils._parallel.Parallel` to make it possible to"
                    " propagate the skore configuration of the current thread to"
                    " the joblib workers."
                ),
                UserWarning,
                stacklevel=2,
            )

        with config_context(**config), warnings.catch_warnings():
            warnings.filters = warning_filters
            return self.function(*args, **kwargs)
